from scipy.signal import welch
from scipy.optimize import curve_fit
from sodetlib.util import get_asd, cprint
import sodetlib as sdl
import numpy as np
import matplotlib.pyplot as plt
import os


def noise_model(freq, wl, n, f_knee):
    """
    Crude model for noise modeling

    Args
    ----
    freq : float
        independent variable in function (frequency Hz)
    wl : float
        White-noise level.
    n : float
        Exponent of 1/f^(n/2) component.
    f_knee : float
        Frequency at which white noise = 1/f^n component
    Returns
    -------
    y : float
        dependent variable in function noise in (pA/rtHz)
    """
    y = wl*np.sqrt((f_knee/freq)**n+1)
    return y

def fit_noise_asd(f, Axx, wl_f_range=(10,30), p0=None):
    """
    Return model fit for a ASD.

    Args
    -----

    f : float array
        The frequency information.
    Axx : float array
        The power spectral data.
    wl_f_range: float tuple
        tuple contains (f_low, f_high), f_high is used to determine the
        range of the ASD to fit to the noise model and if fit fails then the
        white noise is calculated as the median between f_low and f_high.
    p0 : float array or None, optional, default None
        Initial guess for fitting ASDs.  If None, sets to ``p0=[100.0,0.5,0.01]``
        which corresponds to: 

          - white-noise level in pA/rtHz
          - exponent of 1/f^n component
          - knee
          - frequency in Hz

    Returns
    -------

    popt : float array
        The fit parameters - [white_noise_level, n, f_knee].
    """
    if p0 is None:
        p0 = [100.,0.5,0.01]

    bounds_low = [0.,0.,0.] # constrain 1/f^n to be red spectrum
    bounds_high = [np.inf,np.inf,np.inf]
    bounds = (bounds_low,bounds_high)

    fit_idxmax = np.nanargmin(np.abs(f-wl_f_range[1]))
    try:
        popt, pcov = curve_fit(noise_model, f[1:fit_idxmax], Axx[1:fit_idxmax],
                                p0=p0, bounds=bounds)
    except Exception:
        idxmin = np.nanargmin(np.abs(f-wl_f_range[0]))
        wl = np.median(Axx[idxmin:fit_idxmax])
        popt = [wl, np.nan, np.nan]
        pcov = None
    return popt

def get_noise_params(am, wl_f_range=(10,30),
                     fit=False, nperdecade=10, **asd_args):
    """
    Function to calculate the ASD from an axis manager and then calculate the
    white noise, and fknee (and n-index of 1/f^{n/2} if fit=True). The fit=True
    option is a lot slower.

    Args
    ----
    am: AxisManager
        axis manager loaded using G3tSmurf with timestamps and signal keys.
    wl_f_range: float tuple
        tuple contains (f_low, f_high), if fit=True see `fit_noise_ASD`.
        The white noise is calculated as the median of the ASD (pA/rtHz)
        between f_low and f_high.
    fit: bool
        if true will fit the ASD using `fit_noise_ASD` function
    nperdecade: int
        number of bins per decade (i.e. between 0.01 to 0.1 or 0.1 to 1) to
        use to average the ASD over to avoid peaks skewing the search for
        fknee when not using the fit=True option. If nperdecade = 10 for
        example then the bins between 0.01 and 1 would be:
        np.concatenate((np.linspace(0.01,0.1,10),np.linspace(0.1,1,10)[1:]))
    Returns
    -------
    outdict: dict
        dictionary that contains all of the calculated noise parameters by
        channel, band averaged white noise levels, and f, axx ndarrays from
        the calculated ASD the keys are:

        noise_pars: ndarray
            shape is [nchans,3] the 3 items in axis=1 are: 0 = white noise,
            1 = n (1/f index nan if fit=False), and 2 = fknee.
        band_medians: ndarray
            shape is [8,1] median white noise level for each band.
        f: ndarray
            frequency array from welch periodogram
        axx: ndarray
            square root of welch output PSD shape is [nchans,len(f)]
        bincenters: ndarray
            Frequencies where the low frequency spectrum is binned when using
            `Fit = False`, if `Fit=True` then just the array of frequencies
            that can be used to plot the fit results in `lowfn`.
        lowfn: ndarray
            shape = [nchans, len(bincenters)]. Binned noise levels
            (if `Fit=False`) or fit noise levels (if Fit=`True`) for the low
            frequency part of the ASD.
        low_f_10mHz: ndarray
            shape = [nchans].Ratio of the ASD at 10mHz relative to a reference
            low-f component that has a white noise component = 65 pA/rtHz and
            low-f scaling of 1/f^{1/2}.
    """
    f, axx = get_asd(am, **asd_args)
    idx10mHz = np.nanargmin(np.abs(f-0.01))
    nlref_10mHz = 65*np.sqrt((0.1/f[idx10mHz])+1)
    noise_pars = np.zeros((np.shape(axx)[0],3))
    if fit == False:
        #Find white noise
        fmask = (wl_f_range[0] < f) & (f < wl_f_range[1])
        wls = np.median(axx[:, fmask], axis=1)
        band_medians = np.zeros(8)
        for i in range(8):
            m = am.ch_info.band == i
            band_medians[i] = np.median(wls[m])

        #Calculate f_knee
        decades = np.arange(np.floor(np.log10(f[1])),2)
        binedges = np.asarray([])
        for i, dec in enumerate(decades[:-1]):
            if i == 0:
                binedges = np.concatenate(
                    (binedges, np.linspace(10**dec, 10**decades[i+1],
                                           nperdecade)))
            else:
                binedges = np.concatenate(
                    (binedges, np.linspace(10**dec, 10**decades[i+1],
                                           nperdecade)[1:]))
        start_bin = min(np.argsort(np.abs(binedges-f[1]))[0:2])
        binedges = binedges[start_bin:]
        bincenters = binedges[:-1]+np.diff(binedges)/2
        lowfn = np.zeros((len(wls),len(bincenters)))
        for ii,be in enumerate(binedges[:-1]):
            m = (f > be) & (f < binedges[ii+1])
            lowfn[:,ii] = np.median(axx[:,m],axis = 1)
        fknees = [
            bincenters[np.nanargmin(np.abs(lowfn[idx]-wls[idx]*np.sqrt(2)))]
            for idx in range(len(wls))
        ]
        fknees = np.asarray(fknees)
        m = [np.nanmax(lowfn[idx]) < np.sqrt(2)*wls[idx]
             for idx in range(len(wls))]
        fknees[m] = f[1]
        #Set noise_pars to have same structure as when we fit.
        noise_pars[:, 0] = wls
        noise_pars[:, 1] = np.full(len(wls),np.nan)
        noise_pars[:, 2] = fknees
    else:
        bincenters = np.logspace(np.log10(f[1]),np.log10(wl_f_range[1]),30)
        lowfn = np.zeros((len(noise_pars[:,0]),len(bincenters)))
        for i, aaxx in enumerate(axx):
            noise_pars[i,:] = fit_noise_asd(f, aaxx, wl_f_range=wl_f_range)
            lowfn[i,:] = noise_model(bincenters,*noise_pars[i,:])
        band_medians = np.zeros(8)
        for i in range(8):
            m = am.ch_info.band == i
            band_medians[i] = np.median(noise_pars[m,0])
    nl_10mHz_rat = axx[:,idx10mHz]/nlref_10mHz
    outdict = {'noise_pars': noise_pars,
               'bands': am.ch_info.band,
               'channels': am.ch_info.channel,
               'band_medians': band_medians,
               'f': f,
               'axx': axx,
               'bincenters': bincenters,
               'lowfn': lowfn,
               'low_f_10mHz': nl_10mHz_rat}
    return outdict

def plot_band_noise(am, nbins=40, noisedict=None, wl_f_range=(10,30),
                    fit=False, nperdecade=10, show_plot=True, save_plot=False,
                    save_dir=None, **asd_args):
    """
    Makes a summary plot w/ subplots per band of a histogram of the white noise
    levels and another plot with histograms of the fknees. If an axis AxisManager
    is passed without a noisedict then `get_noise_params` will be called to
    generate a noisedict otherwise those parameters will be skipped.

    Args
    ----
    am: AxisManager
        axis manager loaded using G3tSmurf with timestamps and signal keys
        to be analyzed.
    nbins: int
        number of bins in the histograms.
    noisedict: dict
        dictionary returned by `get_noise_params` see that docstring for
        details on dictionary keys.
    wl_f_range: float tuple
        tuple contains (f_low, f_high), if fit=True see `fit_noise_asd`.
        The white noise is calculated as the median of the ASD (pA/rtHz)
        between f_low and f_high.
    fit: bool
        if true will fit the ASD using `fit_noise_asd` function
    nperdecade: int
        used to calculate fknee, see `get_noise_params` doc string.
    show_plot: bool
        plot only displayed if true.
    save_plot: bool
        plot only saved if true. If true then `save_dir` is required to
        save properly otherwise will just return without saving.
    save_dir: str
        directory where plots are saved. Required if `save_plot` is True.
    Returns
    -------
    fig_wnl: `matplotlib.figure.Figure`
        matplotlib figure object for white noise plot
    axes_wnl: `matplotlib.axes.Axes`
        matplotlib axes object for white noise plot
    fig_fk: `matplotlib.figure.Figure` 
        matplotlib figure object for fknee plot
    axes_fk: `matplotlib.axes.Axes`
        matplotlib axes object for fknee plot
    """
    if save_plot:
        if save_dir==None:
            raise ValueError('save_dir must be provided to save plots.')
    bands = am.ch_info.band
    ctime = int(am.timestamps[0])

    if noisedict == None:
        noisedict = get_noise_params(am, wl_f_range=wl_f_range, fit=fit,
                                      nperdecade=nperdecade, **asd_args)
    wls = noisedict['noise_pars'][:,0]
    fknees = noisedict['noise_pars'][:,2]
    band_medians = noisedict['band_medians']

    #Only turns off interactive mode inside with so that won't effect setting.
    try:
        isinteractive = plt.isinteractive
        plt.ioff()

        #Plot white noise histograms
        fig_wnl, axes_wnl = plt.subplots(4, 2, figsize=(16, 8),
                                 gridspec_kw={'hspace': 0})
        fig_wnl.patch.set_facecolor('white')
        bins = np.logspace(1, 4, nbins)
        max_bins = 0

        for b in range(8):
            ax = axes_wnl[b % 4, b // 4]
            m = bands == b
            x = ax.hist(wls[m], bins=bins)
            text  = f"Median: {band_medians[b]:0.1f} pA/rtHz\n"
            text += f"Chans pictured: {np.sum(x[0]):0.0f}"
            ax.text(0.7, .7, text, transform=ax.transAxes)
            ax.axvline(np.median(wls[m]), color='red')
            max_bins = max(np.max(x[0]), max_bins)
            ax.set(xscale='log', ylabel=f'Band {b}')

        axes_wnl[0][0].set(title="AMC 0")
        axes_wnl[0][1].set(title="AMC 1")
        axes_wnl[-1][0].set(xlabel="White Noise (pA/rt(Hz))")
        axes_wnl[-1][1].set(xlabel="White Noise (pA/rt(Hz))")
        for _ax in axes_wnl:
            for ax in _ax:
                ax.set(ylim=(0, max_bins * 1.1))
        plt.suptitle(
            f'Total yield {len(wls)}, Overall median noise {np.nanmedian(wls):0.1f} pA/rtHz')
        if save_plot:
            plt.savefig(os.path.join(save_dir,
                                    f'{ctime}_white_noise_summary.png'))
        if show_plot:
            plt.show()
        else:
            plt.close()

        #Plot f_knee histograms
        fig_fk, axes_fk = plt.subplots(4, 2, figsize=(16, 8),
                                 gridspec_kw={'hspace': 0})
        fig_fk.patch.set_facecolor('white')
        bins = np.logspace(np.floor(np.log10(np.min(fknees[fknees>0]))),
                           np.ceil(np.log10(np.max(fknees[fknees>0]))),
                           nbins)
        max_bins = 0

        for b in range(8):
            ax = axes_fk[b % 4, b // 4]
            m = (bands == b) & (fknees > 0)
            x = ax.hist(fknees[m], bins=bins)
            text  = f"Median: {np.median(fknees[m]):0.2f}\n"
            text += f"Chans pictured: {np.sum(x[0]):0.0f}"
            ax.text(0.72, .7, text, transform=ax.transAxes)
            ax.axvline(np.median(fknees[m]), color='red')
            max_bins = max(np.max(x[0]), max_bins)
            ax.set(xscale='log', ylabel=f'Band {b}')

        axes_fk[0][0].set(title="AMC 0")
        axes_fk[0][1].set(title="AMC 1")
        axes_fk[-1][0].set(xlabel="$f_{knee}$ (Hz)")
        axes_fk[-1][1].set(xlabel="$f_{knee}$ (Hz)")
        for _ax in axes_fk:
            for ax in _ax:
                ax.set(ylim=(0, max_bins * 1.1))
        if show_plot:
            plt.show()
        if save_plot:
            plt.savefig(os.path.join(save_dir,f'{ctime}_fknee_summary.png'))
            if not(show_plot):
                plt.close()

        #Plot ASDs
        fig_asd, axes_asd = plt.subplots(4, 2, figsize=(16, 8),
                                 gridspec_kw={'hspace': 0})
        fig_asd.patch.set_facecolor('white')

        min_x, max_x = (1, 0)
        for b in range(8):
            ax = axes_asd[b % 4, b // 4]
            m = bands == b
            med_wl = np.nanmedian(wls[m])
            f_arr = np.tile(noisedict['f'], (sum(m),1))
            x = ax.loglog(f_arr.T, noisedict['axx'][m].T, color='C0', alpha=0.1)
            ax.axhline(med_wl, color='red', alpha=0.6,
                       label=f'Med. WL: {med_wl:.1f} pA/rtHz')
            ax.set(ylabel=f'Band {b}\nASD (pA/rtHz)')
            ax.grid(linestyle='--', which='both')
            ax.legend(loc='upper right')
            min_x = min(ax.get_xlim()[0], min_x)
            max_x = max(ax.get_xlim()[1], max_x)

        axes_asd[0][0].set(title="AMC 0")
        axes_asd[0][1].set(title="AMC 1")
        axes_asd[-1][0].set(xlabel="Frequency (Hz)")
        axes_asd[-1][1].set(xlabel="Frequency (Hz)")
        for _ax in axes_asd:
            for ax in _ax:
                ax.set(xlim=[min_x, max_x], ylim=[1, 5e3])
        if save_plot:
            plt.savefig(os.path.join(save_dir,
                                    f'{ctime}_band_asds.png'))
        if show_plot:
            plt.show()
        else:
            plt.close()

    finally:
        if isinteractive:
            plt.ion()
    return fig_wnl, axes_wnl, fig_fk, axes_fk, fig_asd, axes_asd

def plot_channel_noise(am, rc, save_dir=None, noisedict=None, wl_f_range=(10,30),
                       fit=False, show_plot=False, save_plot=False, nperdecade=10,
                       plot1overfregion=False, **asd_args):
    """
    Function for plotting the tod and psd with white noise and fknee identified
    for a single channel.

    Args
    ----
    am: `sotodlib.core.AxisManager`
        axis manager loaded using G3tSmurf with timestamps and signal keys
        to be analyzed.
    rc: int
        Readout channel (i.e. index of am.signal) to plot.
    noisedict: dict
        dictionary returned by `get_noise_params` see that docstring for
        details on dictionary keys.
    wl_f_range: float tuple
        tuple contains (f_low, f_high), if fit=True see `fit_noise_asd`.
        The white noise is calculated as the median of the ASD (pA/rtHz)
        between f_low and f_high.
    fit: bool
        if true will fit the ASD using `fit_noise_asd` function
    show_plot: bool
        plot only displayed if true.
    save_plot: bool
        plot only saved if true.
    nperdecade: int
        used to calculate fknee, see `get_noise_params` doc string.
    plot1overfregion: bool
        if true plots a line and shaded region that represents the SO
        passing low-f requirement (i.e. fknee set by wl = 65pA/rtHz and
        slope must be <= 1/f^{1/2} in the ASD)
    Returns
    -------
    fig: `matplotlib.figure.Figure`
        matplotlib figure object for plot
    axes: `matplotlib.axes.Axes`
        matplotlib axes object for plot
    """
    if (save_plot) & (save_dir==None):
        raise ValueError('Must provide save path, exiting.')

    if noisedict == None:
        noisedict = get_noise_params(am, wl_f_range=wl_f_range, fit=fit,
                                     nperdecade=nperdecade, **asd_args)
    f,axx = noisedict['f'],noisedict['axx']
    noise_pars = noisedict['noise_pars']

    #Only turns off in with so that doesn't effect outside setting.
    try:
        isinteractive = plt.isinteractive
        plt.ioff()
        props = dict(boxstyle='round', facecolor='wheat', alpha=0.5)

        plt.figure()
        fig, axes = plt.subplots(2, 1)
        ax1 = axes[0]
        ax1.plot((am.timestamps-am.timestamps[0]),am.signal[rc])
        ax1.set_xlabel('Elapsed Time [sec]',fontsize = 14)
        ax1.set_ylabel('Signal [pA]',fontsize = 14)
        band = am.ch_info.band[rc]
        chan = am.ch_info.channel[rc]
        ttlstr = f'Band {band}, Channel {chan}'
        ax1.set_title(ttlstr,fontsize = 18)
        ax2 = axes[1]
        ax2.loglog(f,axx[rc],color = 'grey',alpha = 0.7)
        ax2.axhline(noise_pars[rc,0], xmin=0.5, xmax=1, lw=2, ls=':',
                    color='k')
        wl_fknee = axx[rc,np.nanargmin(np.abs(f-noise_pars[rc,2]))]
        ax2.plot(noise_pars[rc,2], wl_fknee, '*', markersize=8, color='green')
        if plot1overfregion:
            #plt.axhline(np.sqrt(2)*noise_pars[rc,0],color = 'C0')
            #plt.axvline(noise_pars[rc,2])
            plt.plot(noisedict['bincenters'],
                     noisedict['lowfn'][rc],'.',lw = 3,color = 'r')
            l1 = noise_pars[rc,0]*np.ones(len(f[f<=0.1]))
            with np.errstate(divide='ignore'):
                l2 = 65*np.sqrt(0.1/f[f<=0.1])
            ax2.plot(f[f<=0.1],l2,'--',color = 'C1')
            ax2.fill_between(f[f<=0.1],l2,l1,color = 'wheat',alpha = 0.3)
        text = f'White Noise: {np.round(float(noise_pars[rc,0]),1)} pA/rtHz\n'
        text += 'f$_{knee}$: '+f'{np.round(noise_pars[rc,2],4)} Hz'
        ax2.text(0.03, 0.1, text, bbox=props, transform=ax2.transAxes)
        ax2.set_xlabel('Frequency [Hz]', fontsize=14)
        ax2.set_ylabel('ASD [pA/rtHz]', fontsize=14)

        if save_plot:
            ctime = int(am.timestamps[0])
            plt.savefig(os.path.join(save_dir,
                                     f'{ctime}_b{band}c{chan}_noise.png'))
        if show_plot:
            plt.show()
        else:
            plt.close()

    finally:
        if isinteractive:
            plt.ion()
    return fig, axes

@sdl.set_action()
def take_noise(S, cfg, acq_time=30, plot_band_summary=True, nbins=40,
               show_plot=True, save_plot=True, plotted_rchans=None,
               wl_f_range=(10,30), fit=False,
               nperdecade=10, plot1overfregion=False, save_dir=None,
               g3_tag=None,
               **asd_args):
    """
    Streams data for specified amount of time and then calculated the ASD and
    calculates the white noise levels and fknees for all channels. Optionally
    the band medians of the fitted parameters can be plotted and/or individual
    channel plots of the TOD and ASD with white noise and fknee called out.

    Args
    ----

    S : `pysmurf.client.base.smurf_control.SmurfControl`
        pysmurf control object
    cfg : `sodetlib.det_config.DetConfig`
        detconfig object
    acq_time : float
        acquisition time for the noise timestream.
    plot_band_summary : bool
        if true will plot band summary of white noise and  fknees.
    show_plot : bool
        if true will display plots.
    plotted_rchans : int list
        list of readout channels (i.e. index of am.signal) to make channel
        plots for.
    wl_f_range : float tuple
        tuple contains (f_low, f_high), if fit=True see `fit_noise_asd`.
        The white noise is calculated as the median of the ASD (pA/rtHz)
        between f_low and f_high.
    fit : bool
        if true will fit the ASD using `fit_noise_asd` function
    nperdecade : int
        used to calculate fknee, see `get_noise_params` doc string.
    plot1overfregion : bool
        if true plots a line and shaded region that represents the SO
        passing low-f requirement (i.e. fknee set by wl = 65pA/rtHz and
        slope must be <= 1/f^{1/2} in the ASD)
    g3_tag: string, optional
        Tag to be attached to g3 stream 
    
    Returns
    -------
    am: AxisManager
        AxisManager from the timestream acquired to calculate noise
        parameters.
    outdict: dict
        dictionary that contains all calculated noise parameters and
        figure and axes objects for all plots generated. The keys are:

            'noisedict':
                dictionary returned by `get_noise_params` see that doc string
            'fig_wnl':
                matplotlib figure object for white noise band summary plot
                Only returned if plot_band_summary is True.
            'axes_wnl':
                matplotlib axes object for white noise band summary plot
                Only returned if plot_band_summary is True.
            'fig_fk':
                matplotlib figure object for fknee band summary plot
                Only returned if plot_band_summary is True.
            'axes_fk':
                matplotlib axes object for fknee band summary plot
                Only returned if plot_band_summary is True.
            'channel_plots':
                nested dictionary that has a key for each readout channel in the
                plotted_rchans list and contains a matplotlib figure and axis for
                each readout channel. Only returned if plot_channel_noise is True.
    """
    if save_dir is None:
        save_dir = S.plot_dir

    sid = sdl.take_g3_data(S, acq_time, tag=g3_tag, subtype='noise')
    am = sdl.load_session(cfg.stream_id, sid, base_dir=cfg.sys['g3_dir'])
    ctime = int(am.timestamps[0])
    noisedict = get_noise_params(am, wl_f_range=wl_f_range, fit=fit,
                                 nperdecade=nperdecade, **asd_args)

    sdl.set_session_data(S, 'noise', {
        'band_medians': noisedict['band_medians']
    })

    outdict = noisedict.copy()
    outdict['sid'] = sid
    outdict['meta'] = sdl.get_metadata(S, cfg)
    if plot_band_summary:
        fig_wnl, axes_wnl, fig_fk, axes_fk, fig_asd, axes_asd = plot_band_noise(
            am, nbins=nbins, noisedict=noisedict, show_plot=show_plot,
            save_plot=False, save_dir=save_dir, **asd_args)
        if save_plot:
            savename = os.path.join(save_dir, f'{ctime}_white_noise_summary.png')
            fig_wnl.savefig(savename)
            S.pub.register_file(savename, 'take_noise', plot=True)
            savename = os.path.join(save_dir, f'{ctime}_fknee_summary.png')
            fig_fk.savefig(savename)
            S.pub.register_file(savename, 'take_noise', plot=True)
            savename = os.path.join(save_dir, f'{ctime}_asd_summary.png')
            fig_asd.savefig(savename)
            S.pub.register_file(savename, 'take_noise', plot=True)

    if plotted_rchans is not None:
        outdict['channel_plots'] = {}
        for rc in np.atleast_1d(plotted_rchans):
            outdict['channel_plots'][rc] = {}
            fig, axes = plot_channel_noise(am, rc, save_dir=save_dir,
                                           noisedict=noisedict,
                                           show_plot=show_plot,
                                           save_plot=save_plot,
                                           plot1overfregion=plot1overfregion,
                                           **asd_args)
            outdict['channel_plots'][rc]['fig'] = fig
            outdict['channel_plots'][rc]['axes'] = axes
    fname = os.path.join(S.output_dir, f'{ctime}_take_noise.npy')
    outdict['path'] = fname 
    sdl.validate_and_save(fname, outdict, S=S, cfg=cfg, make_path=False)
    return am, outdict

def plot_noise_all(res, range=(0, 200), text_loc=(0.4, 0.7)):
    """
    Plots the white noise distribution of all bands together.

    Args
    -----
    res : dict
        Result from the ``take_noise`` function
    range : tuple
        Range of the histogram
    text_loc : tuple
        Location to place the textbox containing median and file info. These
        coordinates are in the axis tranform frame with (0, 0) being the bottom
        left and (1, 1) being the top right.
    """
    pars = res['noise_pars']
    wls = pars[:, 0]
    fig, ax = plt.subplots()
    hs = ax.hist(wls, range=range, bins=40)
    wlmed = np.nanmedian(wls)
    ch_pict = int(np.sum(hs[0]))
    ch_tot = len(wls)
    ax.axvline(wlmed, color='red', ls='--')
    txt = '\n'.join([
        f'Median: {wlmed:0.2f} pA/rt(Hz)',
        f'{ch_pict}/{ch_tot} chans pictured',
        f"sid: {res['sid']}",
        f"stream_id: {res['meta']['stream_id']}",
        f"path: {os.path.basename(res['path'])}"
    ])
    ax.text(*text_loc, txt, transform=ax.transAxes,
            bbox=dict(facecolor='wheat', alpha=0.7))
    ax.set_xlabel("White Noise (pA/rt(Hz))")
    return fig, ax



import time
import numpy as np
import sodetlib as sdl
from sodetlib.operations.tracking import relock_tracking_setup
from sodetlib.operations import uxm_setup, bias_steps

import matplotlib.pyplot as plt
import os
if not os.environ.get('NO_PYSMURF', False):
    from pysmurf.client.base.smurf_control import SmurfControl



@sdl.set_action()
def reload_tune(S, cfg, bands, setup_notches=False,
                new_master_assignment=False, tunefile=None, update_cfg=True):
    """
    Reloads an existing tune, runs setup-notches and serial grad descent
    and eta scan.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        Det config instance
    bands : list, optional
        List of bands to run
    setup_notches : bool
        Whether to run setup_notches
    new_master_assignment : bool
        Whether to create a new master assignment
    tunefile : str
        Tunefile to load. Defaults to the tunefile in the device cfg.
    update_cfg : bool
        If true, will update the device cfg with the tunefile created by
        setup notches.
    """

    if tunefile is None:
        tunefile = cfg.dev.exp['tunefile']

    S.load_tune(tunefile)

    for band in bands:
        sdl.stop_point(S)
        bcfg = cfg.dev.bands[band]
        S.set_att_uc(band, bcfg['uc_att'])
        S.set_att_dc(band, bcfg['dc_att'])
        sdl.pub_ocs_log(S, f"Relocking tune: Band {band}")
        if setup_notches:
            S.log(f"Setup notches, new_master_assignment={new_master_assignment}")
            S.setup_notches(band, new_master_assignment=new_master_assignment)
        else:
            S.relock(band)

    if setup_notches and update_cfg:
        # Update tunefile
        cfg.dev.update_experiment({'tunefile': S.tune_file}, update_file=True)

    run_grad_descent_and_eta_scan(S, cfg, bands=bands, update_tune=False)

    return True, None


@sdl.set_action()
def run_grad_descent_and_eta_scan(
    S, cfg, bands=None, update_tune=False, force_run=False, max_iters=None,
    gain=None): 
    """
    This function runs serial gradient and eta scan for each band.
    Critically, it pulls in gradient descent tune parameters from the device
    config and uses them to setup the grad descent operation before running.

    Args
    ----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        Det config instance
    bands : list, optional
        List of bands to run on. Defaults to all 8
    update_tune : bool
        If this is set to True, the new resonance frequency and eta parameters
        will be loaded into the smurf tune, and a new tunefile will be written
        based on the new measurements.
    force_run : bool
        If True, will reset the ``etaScanInProgress`` variable to force it
        to re-run. This might be necessary if serial gradient descent failed out.
    """

    if bands is None:
        bands = cfg.dev.exp['active_bands']
    bands = np.atleast_1d(bands)

    for b in bands:
        sdl.stop_point(S)
        in_progress_reg = S._cryo_root(b) + 'etaScanInProgress'
        if S._caget(in_progress_reg):
            if force_run:
                S._caput(in_progress_reg, 0)
            else:
                raise RuntimeError(
                    "Failed during grad descent -- scan already in progress"
                )

        bcfg = cfg.dev.bands[b]

        if max_iters is None:
            max_iters = bcfg['gradientDescentMaxIters']
        if gain is None:
            gain = bcfg['gradientDescentGain']

        S.set_gradient_descent_step_hz(b, bcfg['gradientDescentStepHz'])
        S.set_gradient_descent_max_iters(b, max_iters)
        S.set_gradient_descent_gain(b, gain)
        S.set_gradient_descent_converge_hz(b, bcfg['gradientDescentConvergeHz'])
        S.set_gradient_descent_beta(b, bcfg['gradientDescentBeta'])

        S.log(f"Running grad descent and eta scan on band {b}")

        init_scale_array = S.get_amplitude_scale_array(b)
        init_center_freq_array = S.get_center_frequency_array(b)
        S.run_serial_gradient_descent(b)

        if (S.get_amplitude_scale_array(b) != init_scale_array).any():
            S.log("Grad descent errored out! Resetting state")
            S.set_amplitude_scale_array(b, init_scale_array)
            S.set_center_frequency_array(b, init_center_freq_array)
            S._caput(in_progress_reg, 0)
            raise RuntimeError(
                "Failed during grad descent -- check smurf-streamer logs"
            )

        S.run_serial_eta_scan(b)

        if update_tune:
            # This basically does the inverse of what's in the relock function
            band_center_mhz = S.get_band_center_mhz(b)
            subband_offset = S.get_tone_frequency_offset_mhz(b)
            center_freq_array = S.get_center_frequency_array(b)
            eta_phase_array = S.get_eta_phase_array(b)
            eta_mag_array = S.get_eta_mag_array(b)
            res_freqs = band_center_mhz + subband_offset + center_freq_array

            for res in S.freq_resp[b]['resonances'].values():
                ch = res['channel']
                f0 = res['freq']
                # Pysmurf does this to ignore channels with bad freqs so I 
                # guess we will too
                for ll, hh in S._bad_mask:
                    if ll < f0 < hh:
                        ch = -1

                res['freq'] = res_freqs[ch]
                res['offset'] = center_freq_array[ch]
                res['eta_phase'] = eta_phase_array[ch]
                res['eta_scaled'] = eta_mag_array[ch]

    if update_tune:
        S.log("Saving tune!")
        S.save_tune()
        cfg.dev.exp['tunefile'] = S.tune_file
        cfg.dev.update_file()


def get_full_band_sweep(S, cfg, band, chan):
    """
    This runs full_band_ampl_sweep to get the resonator response around a single
    channel. This is useful for debugging bad channels.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        Det Config instance
    band : int
        smurf-band of channel
    channel : int
        smurf-channel of channel

    Returns
    ----------
    fs : np.ndarray
        Array of frequencies that were swept over
    resp : np.ndarray
        Complex transmission across the frequency range
    """
    sb = S.get_subband_from_channel(band, chan)
    tone_power = cfg.dev.bands[band]['tone_power']
    center_freq_array = S.get_center_frequency_array(band)
    amp_scale_array = S.get_amplitude_scale_array(band)
    eta_phase_array = S.get_eta_phase_array(band)
    eta_mag_array = S.get_eta_mag_array(band)
    fb_enable = S.get_feedback_enable(band)
    fb_enable_arr = S.get_feedback_enable_array(band)

    try:
        fs, resp = S.full_band_ampl_sweep(band, [sb], tone_power, 2, n_step=256)
    finally:
        S.set_center_frequency_array(band, center_freq_array)
        S.set_amplitude_scale_array(band, amp_scale_array)
        S.set_eta_phase_array(band, eta_phase_array)
        S.set_eta_mag_array(band, eta_mag_array)
        S.set_feedback_enable(band, fb_enable)
        S.set_feedback_enable_array(band, fb_enable_arr)
    return fs, resp


def plot_channel_resonance(S, cfg, band, chan):
    """
    Measures and plots resonator properties for a single smurf channel.
    This will perform a scan around the specified channel, and will plot the
    resonance dip along with where smurf thinks the resonance frequency is.
    This is very useful for debugging why a channel isn't reading out properly.

    Args
    -----
    S : (SmurfControl)
        pysmurf instance
    cfg : (DetConfig)
        DetConfig instance
    band : (int)
        smurf band number
    chan : (int)
        smurf channel number
    """
    res_freq = S.channel_to_freq(band, chan)
    fs, resp = get_full_band_sweep(S, cfg, band, chan)

    fs = fs.ravel() + S.get_band_center_mhz(band)
    resp = resp.ravel()
    fs = fs[np.abs(resp) > 0]
    resp = resp[np.abs(resp) > 0]

    fig, axes = plt.subplots(2, 1, figsize=(10, 10))

    # Res-dip plot
    ax = axes[0]
    ax.plot(fs, np.abs(resp))
    ax = ax.twinx()
    ax.plot(fs, np.unwrap(np.angle(resp)), color='C1')
    ax.axvline(res_freq, color='grey', ls='--')

    # Circ plot
    eta_phase = S.get_eta_phase_array(band)[chan]
    eta = np.exp(1.0j * eta_phase * (2*np.pi) / 360)
    Q = np.real(resp * eta)
    I = -np.imag(resp * eta)

    ax = axes[1]
    ax.axvline(0, color='grey')
    ax.axhline(0, color='grey')
    ax.scatter(I, Q, c=np.abs(resp))
    ax.scatter(np.real(resp), np.imag(resp), c=np.abs(resp), alpha=0.3)

    return fig, axes


@sdl.set_action()
def uxm_relock(
    S, cfg, bands=None, show_plots=False,
    setup_notches=False, new_master_assignment=False, reset_rate_khz=None,
    nphi0=None, skip_setup_amps=False):
    """
    Relocks resonators by running the following steps:

        1. Reset state (all off, disable waveform, etc.)
        2. Set amps and check tolerance
        3. load tune, setup_notches, serial grad descent and eta scan
        4. Tracking setup
        5. Measure noise

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        Det config instance
    bands : list, optional
        List of bands to run on. Defaults to all 8
    setup_notches : bool
        If True, will run setup notches instead of relocking
    new_master_assignment : bool
        Whether to run setup_notches with new_master_assignment.
        Defaults to False
    reset_rate_khz : float, optional
        Flux Ramp Reset Rate to set (kHz), defaults to the value in the dev cfg
    nphi0 : int, optional
        Number of phi0's to ramp through. Defaults to the value that was used
        during setup.
    show_plots : bool
        If True will show plots
    skip_setup_amps : bool
        If True will skip amplifier setup.

    Returns
    --------
    summary : dict
        Dictionary containing a summary of all run operations.
    """
    if bands is None:
        bands = cfg.dev.exp['active_bands']
    bands = np.atleast_1d(bands)

    exp = cfg.dev.exp

    #############################################################
    # 1. Reset to known state
    #############################################################
    S.all_off()  # Turn off Flux ramp, tones, and biases
    S.set_rtm_arb_waveform_enable(0)
    S.set_filter_disable(0)
    S.set_downsample_factor(exp['downsample_factor'])
    if exp['coupling_mode'] == 'dc':
        S.set_mode_dc()
    else:
        S.set_mode_ac()

    for band in bands:
        band_cfg = cfg.dev.bands[band]
        S.set_synthesis_scale(band, exp['synthesis_scale'])
        S.set_att_uc(band, band_cfg['uc_att'])
        S.set_att_dc(band, band_cfg['dc_att'])
        S.set_band_delay_us(band, band_cfg['band_delay_us'])
        S.amplitude_scale[band] = band_cfg['tone_power']

    summary = {}
    summary['timestamps'] = []
    #############################################################
    # 2. Setup amps
    #############################################################
    sdl.stop_point(S)
    if not skip_setup_amps:
        summary['timestamps'].append(('setup_amps', time.time()))
        sdl.set_session_data(S, 'timestamps', summary['timestamps'])
        success, summary['amps'] = uxm_setup.setup_amps(S, cfg)

        if not success:
            return False, summary
    else:
        print("Skipping amp setup")

    #############################################################
    # 3. Load tune
    #############################################################
    sdl.stop_point(S)
    summary['timestamps'].append(('load_tune', time.time()))
    sdl.set_session_data(S, 'timestamps', summary['timestamps'])
    success, summary['reload_tune'] = reload_tune(
        S, cfg, bands, setup_notches=setup_notches,
        new_master_assignment=new_master_assignment)

    sdl.set_session_data(S, 'reload_tune', summary['reload_tune'])
    if not success:
        return False, summary

    #############################################################
    # 4. Tracking Setup
    #############################################################
    sdl.stop_point(S)
    summary['timestamps'].append(('tracking_setup', time.time()))
    sdl.set_session_data(S, 'timestamps', summary['timestamps'])

    # Ensure feedback is enabled
    for b in bands:
        S.set_feedback_enable(b, 1)

    tr = relock_tracking_setup(
        S, cfg, bands, show_plots=show_plots,
        reset_rate_khz=reset_rate_khz, nphi0=nphi0
    )
    summary['tracking_setup_results'] = tr

    #############################################################
    # 5. Noise
    #############################################################
    sdl.stop_point(S)
    summary['timestamps'].append(('noise', time.time()))
    sdl.set_session_data(S, 'timestamps', summary['timestamps'])
    am, summary['noise'] = sdl.noise.take_noise(S, cfg, 30,
                                                show_plot=show_plots,
                                                save_plot=True,
                                                g3_tag='post_relock')
    summary['noise']['am'] = am

    sdl.set_session_data(S, 'noise', {
        'band_medians': summary['noise']['band_medians']
    })

    #############################################################
    # 6. Bias group mapping
    #############################################################
    sdl.stop_point(S)
    summary['timestamps'].append(('bg_map', time.time()))
    sdl.set_session_data(S, 'timestamps', summary['timestamps'])

    bsa = bias_steps.take_bgmap(S, cfg,
                                show_plots=show_plots)
    summary['bg_map'] = bsa

    summary['timestamps'].append(('end', time.time()))
    sdl.set_session_data(S, 'timestamps', summary['timestamps'])

    return True, summary



import numpy as np
import time
import sodetlib as sdl
from sodetlib.operations import tracking

import matplotlib.pyplot as plt


def find_gate_voltage(S, target_Id, amp_name, vg_min=-2.0, vg_max=0,
                      max_iter=50, wait_time=0.5, id_tolerance=0.2):
    """
    Scans through bias voltage for hemt or 50K amplifier to get the correct
    gate voltage for a target current.

    Args
    -----
    S: pysmurf.client.SmurfControl
        PySmurf control object
    target_Id: float
        Target amplifier current
    vg_min: float
        Minimum allowable gate voltage
    vg_max: float
        Maximum allowable gate voltage
    amp_name: str
        Name of amplifier. Must be one of ['hemt', 'hemt1', 'hemt2', '50k', '50k1', '50k2'].
    max_iter: int, optional
        Maximum number of iterations to find voltage. Defaults to 30.
    wait_time : float
        Time to wait after setting the voltage at each step. Defaults
        to 0.5 sec
    id_tolerance : float
        Max difference between target drain current and actual drain currents
        for this to be considered success (mA). Defaults to 0.2 mA.

    Returns
    --------
    success : bool
        Returns a boolean signaling whether voltage scan has been successful.
        The set voltages can be read with S.get_amplifier_biases().
    """
    all_amps = S.C.list_of_c02_and_c04_amps
    if amp_name not in all_amps:
        raise ValueError(f"amp_name must be one of {all_amps}")

    for _ in range(max_iter):
        amp_biases = S.get_amplifier_biases()
        Vg = amp_biases[f"{amp_name}_gate_volt"]
        Id = amp_biases[f"{amp_name}_drain_current"]
        delta = target_Id - Id

        S.log(delta)
        # Check if Id is within tolerance
        if np.abs(delta) < id_tolerance:
            return True

        if 'hemt' in amp_name:
            step = np.sign(delta) * (0.1 if np.abs(delta) > 1.5 else 0.01)
        else:
            step = np.sign(delta) * (0.01 if np.abs(delta) > 1.5 else 0.001)

        Vg_next = Vg + step
        if not (vg_min < Vg_next < vg_max):
            S.log(f"Gate voltage adjustment would go out of range ({vg_min}, {vg_max}). "
                  f"Unable to change {amp_name}_drain_current to desired value", False)
            return False

        if 'hemt' in amp_name:
            S.set_amp_gate_voltage(amp_name, Vg_next, override=True)
        else:
            S.set_amp_gate_voltage(amp_name, Vg_next, override=True)

        time.sleep(wait_time)

    S.log(f"Max allowed Vg iterations ({max_iter}) has been reached. "
          f"Unable to get target drain current for {amp_name}.", False)

    return False

@sdl.set_action()
def setup_amps(S, cfg, update_cfg=True, enable_300K_LNA=True):
    """
    Initial setup for 50k and hemt amplifiers. For C04/C05 cryocards, will first
    check if the drain voltages are set. Then checks if drain
    currents are in range, and if not will scan gate voltage to find one that
    hits the target current. Will update the device cfg if successful.

    The following parameters can be modified in the device cfg, where {amp}
    is one of ['hemt', 'hemt1', 'hemt2', '50k', '50k1', '50k2']:

        exp:
         - amp_enable_wait_time (float): Seconds to wait after enabling amps
           before scanning gate voltages
         - amp_{amp}_drain_current (float): Target drain current (mA)
         - amp_{amp}_drain_current_tolerance (float): Tolerance for drain current (mA)
         - amp_{amp}_drain_volt (float) : Drain voltage (V). C04/C05 cryocards only.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        DetConfig instance
    update_cfg : bool
        If true, will update the device cfg and save the file.
    enable_300K_LNA:
        If true, will turn on the 300K LNAs.
    """
    sdl.pub_ocs_log(S, "Starting setup_amps")

    exp = cfg.dev.exp

    # Determine cryocard rev
    major, minor, patch = S.C.get_fw_version()
    if major == 4:
        cc_rev = 'c04'
        amp_list = S.C.list_of_c04_amps
    elif major == 0:
        if sum((major,minor,patch)) == 0:
               raise ValueError("Error communicatin with cryocard; "
                                + "is it connected?")
        else:
               raise ValueError(f"Unrecognized cryocard firmware version "
                                + "({major},{minor},{patch}).")
    else:
        cc_rev = 'c02'
        amp_list = S.C.list_of_c02_amps
        S.C.write_ps_en(0b11)
        time.sleep(exp['amp_enable_wait_time'])
        if S.C.read_ps_en() != 3:
            raise ValueError("Could not enable amps.")

    amp_list = list(set(amp_list) & set(exp['amps_to_bias']))
    if len(amp_list) == 0:
        raise ValueError(
            f"exp['amps_to_bias']: {exp['amps_to_bias']} contains no valid amps."
        )
    # Data to be passed back to the ocs-pysmurf-controller clients
    summary = {'success': False}
    for amp in amp_list:
        summary[f'{amp}_gate_volt'] = None
        summary[f'{amp}_drain_current'] = None
        summary[amp + '_enable'] = None
        if cc_rev == 'c04':
            summary[f'{amp}_drain_volt'] = None

    amp_biases = S.get_amplifier_biases()

    # For C04, first check drain voltages
    if cc_rev == 'c04':
        for amp in amp_list:
            Vd = amp_biases[f"{amp}_drain_volt"]
            if Vd != exp[f"amp_{amp}_drain_volt"]:
                S.set_amp_drain_voltage(amp, exp[f"amp_{amp}_drain_volt"]) 

    # Check drain currents / scan gate voltages
    delta_drain_currents = dict()
    for amp in amp_list:
        delta_Id = np.abs(amp_biases[f"{amp}_drain_current"] - exp[f"amp_{amp}_drain_current"])
        if delta_Id > exp[f'amp_{amp}_drain_current_tolerance']:
            S.log(f"{amp} current not within tolerance, scanning for correct gate voltage")
            S.set_amp_gate_voltage(amp, exp[f'amp_{amp}_init_gate_volt'],override=True)
            success = find_gate_voltage(
                S, exp[f"amp_{amp}_drain_current"], amp, wait_time=exp['amp_step_wait_time'],
                id_tolerance=exp[f'amp_{amp}_drain_current_tolerance']
            )
            if not success:
                sdl.pub_ocs_log(S, f"Failed determining {amp} gate voltage")
                sdl.set_session_data(S, 'setup_amps_summary', summary)
                S.C.write_ps_en(0)
                return False, summary
    # Turn on 300K LNAs
    if enable_300K_LNA:
        S.C.write_optical(0b11)
        
    # Update device cfg
    biases = S.get_amplifier_biases()
    if update_cfg:
        for amp in amp_list:
            cfg.dev.update_experiment({f'amp_{amp}_gate_volt': biases[f'{amp}_gate_volt']},
                                      update_file=True)

    summary = {'success': True, **biases}
    sdl.set_session_data(S, 'setup_amps_summary', summary)
    return True, summary


@sdl.set_action()
def setup_phase_delay(S, cfg, bands, update_cfg=True):
    """
    Runs estimate phase delay and updates the device cfg with the results.
    This will run with the current set of attenuation values.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        DetConfig instance
    uc_att : int
        UC atten to use for phase-delay estimation
    dc_att : int
        DC atten to use for phase-delay estimation
    update_cfg : bool
        If true, will update the device cfg and save the file.
    """
    sdl.pub_ocs_log(S, f"Estimating phase delay for bands {bands}")

    summary = {
        'bands': [],
        'band_delay_us': [],
    }
    for b in bands:
        summary['bands'].append(int(b))
        band_delay_us, _ = S.estimate_phase_delay(b, make_plot=True, show_plot=False)
        band_delay_us = float(band_delay_us)
        summary['band_delay_us'].append(band_delay_us)
        if update_cfg:
            cfg.dev.bands[b].update({
                'band_delay_us': band_delay_us
            })

    if update_cfg:
        cfg.dev.update_file()

    sdl.set_session_data(S, 'setup_phase_delay', summary)
    return True, summary


def estimate_uc_dc_atten(S, cfg, band, update_cfg=True, tone_power=None):
    """
    Provides an initial estimation of uc and dc attenuation for a band in order
    to get the band response at a particular frequency to be within a certain
    range. The goal of this function is to choose attenuations such that the
    ADC isn't saturated, and the response is large enough to find resonators.

    For simplicity, this searches over the parameter space where uc and dc
    attens are equal, instead of searching over the full 2d space.

    For further optimization of attens with respect to median white noise, run
    the ``optimize_attens`` function from the
    ``sodetlib/operations/optimize.py``
    module.

    The following parameters can be modified in the device cfg:

        bands:
         - tone_power (int): Tone power to use for atten estimation

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    band : int
        Band to estimate attens for
    update_cfg : bool
        If true, will update the device cfg and save the file.  """
    att, step = 15, 7
    sb = S.get_closest_subband(-200, band)

    resp_range = (0.1, 0.4)
    # Just look at 5 subbands after specified freq.
    sbs = np.arange(sb, sb + 5)
    sbs = np.arange(250, 255)
    nread = 2

    success = False
    S.log(f"Estimating attens for band {band}")

    if tone_power is None:
        tone_power = cfg.dev.bands[band]['tone_power']

    while True:
        S.set_att_uc(band, att)
        S.set_att_dc(band, att)

        S.log(f'tone_power: {tone_power}')
        _, resp = S.full_band_ampl_sweep(band, sbs, cfg.dev.bands[band]['tone_power'], nread)
        max_resp = np.max(np.abs(resp))
        S.log(f"att: {att}, max_resp: {max_resp}")

        if resp_range[0] < max_resp < resp_range[1]:
            S.log(f"Estimated atten: {att}")
            success = True
            break

        if max_resp < resp_range[0]:
            if att == 0:
                S.log(f"Cannot achieve resp in range {resp_range}! Try increasing tone power")
                success = False
                break
            att -= step

        elif max_resp > resp_range[1]:
            if att == 30:
                S.log(f"Cannot achieve resp in range {resp_range}! Try decreasing tone power")
                success = False
                break
            att += step
        step = int(np.ceil(step / 2))

    if success and update_cfg:
        cfg.dev.update_band(band, {
            'uc_att': att,
            'dc_att': att,
        }, update_file=True)

    return success


@sdl.set_action()
def setup_tune(S, cfg, bands, show_plots=False, update_cfg=True):
    """
    Find freq, setup notches, and serial gradient descent and eta scan

    The following parameters can be modified in the device cfg:

        exp:
         - res_amp_cut (float): Amplitude cut for peak-finding in find-freq
         - res_grad_cut (float): Gradient cut for peak-finding in find-freq
        bands:
         - tone_power (int): Tone power to use for atten estimation

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        DetConfig instance
    show_plots : bool
        If true, will show find_freq plots. Defaults to False
    update_cfg : bool
        If true, will update the device cfg and save the file.
    """
    bands = np.atleast_1d(bands)
    sdl.pub_ocs_log(S, f"Starting setup_tune for bands {bands}")

    exp = cfg.dev.exp

    summary = {}

    for band in bands:
        sdl.pub_ocs_log(S, f"Find Freq: band {band}")
        bcfg = cfg.dev.bands[band]
        S.find_freq(band, tone_power=bcfg['tone_power'], make_plot=True,
                    save_plot=True, show_plot=show_plots,
                    amp_cut=exp['res_amp_cut'], 
                    grad_cut=exp['res_grad_cut'])

    for band in bands:
        bcfg = cfg.dev.bands[band]
        sdl.pub_ocs_log(S, f"Setup Notches: band {band}")
        S.setup_notches(band, tone_power=bcfg['tone_power'], new_master_assignment=True)

    for band in bands:
        sdl.pub_ocs_log(S, f"Serial grad descent and eta scan: band {band}")
        S.run_serial_gradient_descent(band)
        S.run_serial_eta_scan(band)

    if update_cfg:
        cfg.dev.update_experiment({'tunefile': S.tune_file}, update_file=True)

    return True, summary


def _find_fixed_tone_freq(S, band, ntone, min_spacing, band_start=-250, band_end=250):
    """
    Place fixed tones in the largest gaps between resonators.

    Args
    ----
    S : SmurfControl
        Pysmurf instance
    band : int
        The band to populate with fixed tones.
    ntone : int
        The number of fixed tones to attempt to place in this band.
    min_spacing : float
        The minimum spacing in MHz between resonators to place a fixed tone in.
    band_start : float
        The lower edge of the band in MHz relative to band center.
    band_end : float
        The upper edge of the band in MHz relative to band center.
    """

    # get resonator and channel frequencies
    res_freq = np.sort([res["freq"] for res in S.freq_resp[band]["resonances"].values()])

    # only consider central part of the band
    band_center = S.get_band_center_mhz(band)
    freq_low = band_center + band_start
    freq_high = band_center + band_end
    res_freq = res_freq[(res_freq > freq_low) & (res_freq < freq_high)]

    # add the band edges
    res_freq = np.concatenate(([freq_low], res_freq, [freq_high]))

    # gaps between resonators
    gaps = np.diff(res_freq)

    # check there are enough for the number of tones
    ntone_per_gap = np.floor(gaps / min_spacing)
    num_gaps = ntone_per_gap.sum()
    if num_gaps < ntone:
        S.log(
            f"Only {num_gaps} gaps greater than {min_spacing} MHz were found in band {band}. "
            "That number of fixed tones will be set."
        )
        ntone = int(num_gaps)
    if ntone == 0:
        return np.array([])
    gaps = gaps[ntone_per_gap > 0]
    res_freq = res_freq[np.where(ntone_per_gap > 0)[0]]  # just the left edge now

    # distribute uniformly in available space, accounting for edge buffer
    gaps_cum = np.cumsum(gaps - min_spacing)  # can place a tone anywhere in this space
    tones = np.linspace(0, gaps_cum[-1], ntone, endpoint=True)
    tone_freqs = np.zeros_like(tones)
    last_edge = -1  # allow for tone on the leading 0 edge
    for i in range(gaps.size):
        # identify which gap they are in
        tones_in_gap = (tones > last_edge) & (tones <= gaps_cum[i])
        # shift back to real frequency
        tone_freqs[tones_in_gap] = (
            tones[tones_in_gap] - last_edge  # position relative to buffer
            + res_freq[i] + min_spacing / 2  # real edge in band + buffer
        )
        last_edge = gaps_cum[i]

    return tone_freqs


@sdl.set_action()
def setup_fixed_tones(
    S, cfg, bands=None, fixed_tones_per_band=8, min_gap=10, tone_power=10, update_cfg=True
    ):
    """
    Identify gaps between resonators and attempt to place a number of fixed tones there.
    These will be saved for every band in the device config. A tune should be available
    before running this action.

    Args
    ----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        DetConfig instance
    bands : list of int
        The bands to set fixed tones in. Get from config by default.
    fixed_tones_per_band : int
        The number of tones to place in each band. Defaults to 8.
    min_gap : float
        The minimum gap size in order to place a fixed tone, in MHz. Defaults to 10.
    tone_power : int
        The tone power to use for the fixed tones. Defaults to 10.
    update_cfg : bool
        Whether to save the fixed tone configuration to the device config. Defaults to True.

    Returns
    -------
    fixed_tones : dict
        Fixed tones configuration, as saved to device config.
    """

    if bands is None:
        bands = cfg.dev.exp['active_bands']

    # ensure we have a tune available
    try:
        _ = S.freq_resp[bands[0]]["resonances"]
    except KeyError as e:
        raise ValueError("No tune is available to set fixed tones from.") from e

    # identify available frequencies
    ft_freq = []
    for b in bands:
        ft_freq += list(_find_fixed_tone_freq(S, b, fixed_tones_per_band, min_gap))

        # turn off any existing fixed tones
        try:
            channels = cfg.dev.bands[b]["fixed_tones"]["channels"]
            amp = S.get_amplitude_scale_array(b)
            amp[channels] = 0.0
            S.set_amplitude_scale_array(b, amp)
            S.log(f"Turned off current fixed tones on band {b}.") 
        except KeyError:
            pass

    # set the fixed tones
    tunefile = cfg.dev.exp.get("tunefile", None)
    fixed_tones = {
        b: {
            "enabled": False,
            "freq_offsets": [],
            "channels": [],
            "tone_power": tone_power,
            "tunefile": tunefile,
        }
        for b in bands
    }
    for fr in ft_freq:
        # this function sets the tone power and disables feedback
        try:
            band, channel = S.set_fixed_tone(fr, tone_power=tone_power)
        except AssertionError as e:
            S.log(f"Failed to assign a fixed tone on frequency {fr} MHz.\n{e}.")
            continue
        # read back the assigned frequency offset
        foff = S.get_center_frequency_mhz_channel(band, channel)
        fixed_tones[band]["freq_offsets"].append(float(foff))
        fixed_tones[band]["channels"].append(int(channel))
        fixed_tones[band]["enabled"] = True

    # update config
    if update_cfg:
        for band, data in fixed_tones.items():
            cfg.dev.update_band(band, {"fixed_tones": data}, update_file=True)

    S.log("Done setting fixed tones.")

    return fixed_tones


@sdl.set_action()
def turn_off_fixed_tones(S, cfg, bands=None):
    """
    Read fixed tones from the device config and set their amplitude to 0.

    Args
    ----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        DetConfig instance
    bands : list of int
        The bands to operate on. Get from config by default.
    """

    if bands is None:
        bands = cfg.dev.exp['active_bands']

    for band in bands:
        # try to read from config
        try:
            if not cfg.dev.bands[band]["fixed_tones"].get("enabled", True):
                S.log(f"Fixed tones are already disabled on band {band}.")
                continue
            channels = cfg.dev.bands[band]["fixed_tones"]["channels"]
        except KeyError as e:
            raise ValueError(
                "No fixed tones present in device config."
            ) from e

        amp = S.get_amplitude_scale_array(band)
        amp[channels] = 0
        S.set_amplitude_scale_array(band, amp)
        cfg.dev.bands[band]["fixed_tones"]["enabled"] = False


@sdl.set_action()
def turn_on_fixed_tones(S, cfg, bands=None):
    """
    Read fixed tones from the device config and ensure the corresponding channels
    are set to the specified amplitude and have feedback disabled.

    Args
    ----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        DetConfig instance
    bands : list of int
        The bands to operate on. Get from config by default.
    """

    if bands is None:
        bands = cfg.dev.exp['active_bands']

    for band in bands:
        # try to read from config
        try:
            if not cfg.dev.bands[band]["fixed_tones"].get("enabled", False):
                S.log(f"Fixed tones are not enabled on band {band}.")
                continue
            foff = cfg.dev.bands[band]["fixed_tones"]["freq_offsets"]
            channels = cfg.dev.bands[band]["fixed_tones"]["channels"]
            tone_power = cfg.dev.bands[band]["fixed_tones"].get("tone_power", None)
            tunefile = cfg.dev.bands[band]["fixed_tones"].get("tunefile", None)
        except KeyError as e:
            raise ValueError(
                "No fixed tones present in device config. Use `setup_fixed_tones` to add some."
            ) from e
        if len(foff) != len(channels):
            raise ValueError(
                "Number of fixed tone channels in device config is different from number of "
                f"frequency offsets for band {band} ({len(channels)} != {len(foff)})."
            )
        cfg_tunefile = cfg.dev.exp.get("tunefile", None)
        if cfg_tunefile is None or tunefile is None:
            # warn but do nothing
            S.log(
                "Turning on fixed tones with no associated tunefile could lead "
                "to conflicts with resonator channels.", S.LOG_ERROR
            )
        elif tunefile != cfg_tunefile:
            raise ValueError(
                "The tunefile has changed since fixed tones were set up. Run `setup_fixed_tones "
                "to find available channels."
            )
        S.log(f"Enabling {len(channels)} fixed tones on band {band}.")

        # set the channel center frequency for each tone
        for df, chan in zip(foff, channels):
            S.set_center_frequency_mhz_channel(band, chan, df)

        # set amplitude and feedback
        amp = S.get_amplitude_scale_array(band)
        feedback = S.get_feedback_enable_array(band)

        amp[channels] = tone_power
        feedback[channels] = 0

        S.set_amplitude_scale_array(band, amp)
        S.set_feedback_enable_array(band, feedback)


@sdl.set_action()
def uxm_setup(S, cfg, bands=None, show_plots=True, update_cfg=True,
              modify_attens=True, skip_phase_delay=False,
              skip_setup_amps=False): 
    """
    The goal of this function is to do a pysmurf setup completely from scratch,
    meaning no parameters will be pulled from the device cfg.

    The following steps will be run:

        1. setup amps
        2. Estimate phase delay
        3. Setup tune
        4. setup tracking
        5. Measure noise

    The following device cfg parameters can be changed to modify behavior:

        exp:
         - downsample_factor (int): Downsample factor to use
         - coupling_mode (str): Determines whether to run in DC or AC mode. Can
           be 'dc' or 'ac'.
         - synthesis_scale (int): Synthesis scale to use
         - amp_enable_wait_time (float): Seconds to wait after enabling amps
           before scanning gate voltages
         - amp_hemt_Id (float): Target drain current for hemt amp (mA)
         - amp_50k_Id (float): Target drain current for 50k amp (mA)
         - amp_hemt_Id_tolerance (float): Tolerance for hemt drain current (mA)
         - amp_50k_Id_tolerance (float): Tolerance for 50k drain current (mA)
         - res_amp_cut (float): Amplitude cut for peak-finding in find-freq
         - res_grad_cut (float): Gradient cut for peak-finding in find-freq
        bands:
         - tone_power (int): Tone power to use for atten estimation

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        DetConfig instance
    show_plots : bool
        If true, will show find_freq plots. Defaults to False
    update_cfg : bool
        If true, will update the device cfg and save the file.
    modify_attens : bool
        If true or attenuations are not set in the device config, will run
        estimate_uc_dc_atten to find a set of attenuations that will work for
        estimate_phase_delay and find_freq. If False and attenuation values
        already exist in the device config, this will use those values.
    skip_phase_delay : bool
        If True, will skip the estimate_phase_delay step
    skip_setup_amps : bool
        If True, will skip amplifier setup
    """
    if bands is None:
        bands = cfg.dev.exp['active_bands']
    bands = np.atleast_1d(bands)

    exp = cfg.dev.exp

    #############################################################
    # 1. Reset to known state
    #############################################################
    S.all_off()  # Turn off Flux ramp, tones, and biases
    S.set_rtm_arb_waveform_enable(0)
    S.set_filter_disable(0)
    S.set_downsample_factor(exp['downsample_factor'])
    if exp['coupling_mode'] == 'dc':
        S.set_mode_dc()
    else:
        S.set_mode_ac()

    for band in bands:
        S.set_synthesis_scale(band, exp['synthesis_scale'])

    summary = {}
    summary['timestamps'] = []

    #############################################################
    # 2. Setup amps
    #############################################################
    if not skip_setup_amps:
        summary['timestamps'].append(('setup_amps', time.time()))
        sdl.set_session_data(S, 'timestamps', summary['timestamps'])

        success, summary['setup_amps'] = setup_amps(S, cfg, update_cfg=update_cfg)
        if not success:
            sdl.pub_ocs_log(S, "UXM Setup failed on setup amps step")
            return False, summary

    #############################################################
    # 3. Estimate Attens
    #############################################################
    summary['timestamps'].append(('estimate_attens', time.time()))
    sdl.set_session_data(S, 'timestamps', summary['timestamps'])
    for band in bands:
        bcfg = cfg.dev.bands[band]
        if modify_attens or (bcfg['uc_att'] is None) or (bcfg['dc_att'] is None):
            success = estimate_uc_dc_atten(S, cfg, band, update_cfg=update_cfg)
            if not success:
                sdl.pub_ocs_log(S, f"Failed to estimate attens on band {band}")
                return False, summary
        else:
            S.set_att_uc(band, bcfg['uc_att'])
            S.set_att_dc(band, bcfg['dc_att'])

    #############################################################
    # 4. Estimate Phase Delay
    #############################################################
    if not skip_phase_delay:
        summary['timestamps'].append(('setup_phase_delay', time.time()))
        sdl.set_session_data(S, 'timestamps', summary['timestamps'])
        success, summary['setup_phase_delay'] = setup_phase_delay(
            S, cfg, bands, update_cfg=update_cfg)
        if not success:
            S.log("UXM Setup failed on setup phase delay step")
            return False, summary

    #############################################################
    # 5. Setup Tune
    #############################################################
    summary['timestamps'].append(('setup_tune', time.time()))
    sdl.set_session_data(S, 'timestamps', summary['timestamps'])
    success, summary['setup_tune'] = setup_tune(
        S, cfg, bands, show_plots=show_plots, update_cfg=update_cfg,)
    if not success:
        S.log("UXM Setup failed on setup tune step")
        return False, summary

    #############################################################
    # 6. Setup Tracking
    #############################################################
    summary['timestamps'].append(('setup_tracking', time.time()))
    sdl.set_session_data(S, 'timestamps', summary['timestamps'])

    # Ensure feedback is enabled
    for b in bands:
        S.set_feedback_enable(b, 1)

    tracking_res = tracking.setup_tracking_params(
        S, cfg, bands, show_plots=show_plots, update_cfg=update_cfg
    )
    summary['tracking_res'] = tracking_res

    #############################################################
    # 7. Noise
    #############################################################
    summary['timestamps'].append(('noise', time.time()))
    sdl.set_session_data(S, 'timestamps', summary['timestamps'])
    _, summary['noise'] = sdl.noise.take_noise(
        S, cfg, 30, show_plot=show_plots, save_plot=True
    )

    summary['timestamps'].append(('end', time.time()))
    sdl.set_session_data(S, 'timestamps', summary['timestamps'])

    return True, summary



"""
Constants useful for general sodetlib ops.
"""
import numpy as np

NBGS = 12
NBANDS = 8
CHANS_PER_BAND= 512
pA_per_phi0 = 9e6
pA_per_rad = pA_per_phi0 / (2*np.pi)

# Max bias voltage is 20 V, and there are 20 bits (about 1.0973e-5)
rtm_bit_to_volt = 20/(2**20)



import numpy as np
import sodetlib as sdl
import time
import matplotlib.pyplot as plt
from sodetlib.operations.iv import IVAnalysis
from sodetlib.operations import bias_steps, iv
np.seterr(all="ignore")


def bias_to_volt_arr(S, cfg, biases, bias_groups=None, overbias=True):
    """
    Biases detectors using an array of voltages.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        det config instance
    biases : np.ndarray
        This must be an array of length 12, with the ith element being the
        voltage bias of bias group i. Note that the bias of any bias group
        that is not active, or not specified in the ``bias_groups`` parameter
        will be ignored.
    bias_groups : list
        List of bias groups to bias. If None, this will default to the active
        bgs specified in the device cfg.
    overbias : bool
        If true, will overbias specified bias lines. If false, will set bias
        voltages without overbiasing detectors.
    """
    if bias_groups is None:
        bias_groups = cfg.dev.exp['active_bgs']
    bias_groups = np.atleast_1d(bias_groups)

    _biases = S.get_tes_bias_bipolar_array()
    for bg in bias_groups:
        _biases[bg] = biases[bg]

    S.log(f"Target biases: ")
    for bg in bias_groups:
        S.log(f"BG {bg}: {biases[bg]:.2f}")

    if overbias:
        sdl.overbias_dets(S, cfg, bias_groups=bias_groups)

    S.set_tes_bias_bipolar_array(biases)
    return


sdl.set_action()
def bias_to_rfrac(S, cfg, rfrac=None, bias_groups=None, iva=None,
                  overbias=True, Rn_range=(5e-3, 12e-3), math_only=False):
    """
    Biases detectors in their superconducting transitions to a specified fraction
    of their normal resistance ("rfrac"). This function will choose TES bias
    voltages for each bias-group that either (a) are the median bias voltage
    achieving the specified rfrac value, or (b) maximizes the number of channels
    that will be placed in the allowable rfrac range.

    Args
    ----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        Detconfig instance
    rfrac : float or 2-tuple
        If float, target rfrac value. If tuple, range of allowable rfracs.
    bias_groups : list, optional
        Bias groups to bias. Defaults to active bgs in device cfg.
    iva : IVAnalysis, optional
        IVAnalysis object. If this is passed, will use it to determine
        bias voltages. Defaults to using value in device cfg.
    overbias_voltage : float
        Voltage to use to overbias detectors
    overbias_wait : float
        Time (sec) to wait at overbiased voltage
    Rn_range : tuple
        A "reasonable" range for the TES normal resistance. This will
        be cut on when determining which IVs should be used to determine
        the optimal bias-voltage.
    math_only : bool
        If this is set, will not actually over-bias voltages, and will
        just return the target biases.

    Returns
    ----------
    biases : np.ndarray
        Array of Smurf bias voltages. Note that this includes all smurf
        bias lines (all 15 of them), but only voltages for the requested
        bias-groups will have been modified.
    """
    if bias_groups is None:
        bias_groups = cfg.dev.exp['active_bgs']
    bias_groups = np.atleast_1d(bias_groups)

    if iva is None:
        iva = IVAnalysis.load(cfg.dev.exp['iv_file'])

    if rfrac is None:
        rfrac = cfg.dev.exp['rfrac']
    rfrac = np.atleast_1d(rfrac)

    if len(rfrac) not in [1,2]:
        raise ValueError(
            f"`rfrac` specified as {rfrac} but must be either float or 2-tuple")
    biases = S.get_tes_bias_bipolar_array()

    Rfrac = (iva.R.T / iva.R_n).T
    if len(rfrac) == 2:
        in_range = (rfrac[0] < Rfrac) & (Rfrac < rfrac[1])

    for bg in bias_groups:
        m = (iva.bgmap == bg)
        m = m & (Rn_range[0] < iva.R_n) & (iva.R_n < Rn_range[1])

        if not m.any():
            continue

        if len(rfrac) == 1:
            target_biases = []
            for rc in np.where(m)[0]:
                target_idx = np.nanargmin(np.abs(Rfrac[rc] - rfrac[0]))
                target_biases.append(iva.v_bias[target_idx])
            biases[bg] = np.median(target_biases)
        else:
            nchans_in_range = np.sum(in_range[m, :], axis=0)
            target_idx = np.nanargmax(nchans_in_range)
            biases[bg] = iva.v_bias[target_idx]

    if math_only:
        return biases

    S.log(f"Target biases: ")
    for bg in bias_groups:
        S.log(f"BG {bg}: {biases[bg]:.2f}")

    if overbias:
        sdl.overbias_dets(S, cfg, bias_groups=bias_groups)

    S.set_tes_bias_bipolar_array(biases)

    return biases


bias_to_rfrac_range = bias_to_rfrac



sdl.set_action()
def biasstep_rebias(
        S, cfg, target_percentage_rn=0.5, bias_groups=None,
        show_plots=True, make_plots=True, g3_tag=None):

    """
    re-bias detectors by taking bias_steps.
    
     Args
    -----
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        det config instance
    target_percentage_rn : float
        this is the target percentage rn for the script to bias to, numbers should be
        between 0.3 and 0.7 for general usage.
    bias_groups : list
        List of bias groups to bias. If None, this will default to the active
        bgs specified in the device cfg.
    make_plots : bool
        If this is set to False, the script will still apply the bias-voltage but will not
        generate plots
    show_plots : bool
        If this is set to False, the script will not show plots
    g3_tag : str
        User settable tag for g3 streams

    Returns
    ----------
    bsa_final :
        analysis result of the last biasstep taken with the re-biased bais voltage
    vbias_estimate_final :
        final bias_voltages calculated and applied by the script
    """
    if bias_groups is None:
        bias_groups = cfg.dev.exp['active_bgs']
    bias_groups = np.atleast_1d(bias_groups)

    bs_kwargs = dict(g3_tag=g3_tag, stream_subtype='biasstep_rebias')

    ## take the initial biasstep
    S.log("taking the first biasstep")
    S.log(f"Initial dc biases {S.get_tes_bias_bipolar_array()}")
    bsa_0 = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)
    if make_plots:
        fig, ax = bias_steps.plot_Rfrac(bsa_0)
        fname = sdl.make_filename(S, 'intial_Rfrac.png', plot=True)
        fig.savefig(fname)
        S.pub.register_file(fname, 'biasstep_rfrac', format='png', plot=True)
        if not show_plots:
            plt.close(fig)
            
            
    ## load IV analysis result so can use dynamic step size
    iva = iv.IVAnalysis.load(bsa_0.meta['iv_file'])
                
    repeat_biasstep = False
    ## check for script failue
    for bl in bias_groups:
        mask_bg = np.where(bsa_0.bgmap==bl)[0]
        per_bl_R0 = bsa_0.R0[mask_bg]
        if np.isnan(np.nanmedian(per_bl_R0)):
            ## the yield is very low, could it be the biasstep analysis script failed? repeat again
            repeat_biasstep = True
            S.log("the biasstep script might have failed, going to repeat the measurement")

    if repeat_biasstep:
        S.log("retaking biasstep due to failure")
        bsa_0 = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)
        repeat_biasstep = False
    
    percentage_rn_0 = bsa_0.R0/bsa_0.R_n_IV
    ## add in checks if detectors are SC  
    bg_overbias_needed = []
    overbias_needed = False
    for bl in bias_groups:
        mask_bg = np.where(bsa_0.bgmap==bl)[0]
        per_bl_percentage_rn_0 = percentage_rn_0[mask_bg]
        ## mask for SC
        mask_sc = np.where((per_bl_percentage_rn_0 < 0.1) & (per_bl_percentage_rn_0 > -0.1))
        ## if more than 10 percent of the detectors are SC
        if len(mask_sc[0]) > 0.1*len(mask_bg):
            bg_overbias_needed.append(bl)
            overbias_needed = True

    ## add in check if detectors are normal, hence we should increase the drop step
    bg_detectors_normal = []
    drop_from_normal = False
    for bl in bias_groups:
        mask_bg = np.where(bsa_0.bgmap==bl)[0]
        per_bl_percentage_rn_0 = percentage_rn_0[mask_bg]
        ## mask for normal
        mask_normal = np.where((per_bl_percentage_rn_0 > 0.9) | (per_bl_percentage_rn_0 < -0.9))
        ## if more than half of the detectors are normal
        if len(mask_normal[0]) > 0.5*len(mask_bg):
            bg_detectors_normal.append(bl)   
            drop_from_normal = True


    if overbias_needed:
        S.log(f"some BL: {bg_overbias_needed} are stuck in SC, going to over bias them, this takes about 5mins")
        previous_dc_biases = S.get_tes_bias_bipolar_array()
        
        sdl.overbias_dets(S, cfg, bias_groups=bg_overbias_needed)
        ##force wait for UFM to reach equalibrium 
        sleep_time = cfg.dev.exp["overbias_sleep_time_sec"]
        time.sleep(sleep_time)
        safe_dc_biases = previous_dc_biases.copy()
        for replace_bg in bg_overbias_needed:
            try:
                safe_dc_biases[replace_bg] = cfg.dev.bias_groups[replace_bg]["testbed_100mK_bias_voltage"]
            except KeyError:
                mask_bg = np.where(bsa_0.bgmap==replace_bg)[0]
                v_norm_bl = np.nanmedian(iva.v_bias[iva.idxs[[mask_bg], 1]])
                safe_dc_biases[replace_bg] = v_norm_bl
                
            
        S.set_tes_bias_bipolar_array(safe_dc_biases)  
        bsa_0 = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)
        if make_plots:
            fig, ax = bias_steps.plot_Rfrac(bsa_0)
            fname = sdl.make_filename(S, 'post_overbiasing_Rfrac.png', plot=True)
            fig.savefig(fname)
            S.pub.register_file(fname, 'biasstep_rfrac', format='png', plot=True)
            if not show_plots:
                plt.close(fig)
        
        ## add in check if detectors are normal now
        percentage_rn_0 = bsa_0.R0/bsa_0.R_n_IV
        drop_from_normal = False
        for bl in bg_overbias_needed:
            mask_bg = np.where(bsa_0.bgmap==bl)[0]
            per_bl_percentage_rn_0 = percentage_rn_0[mask_bg]
            ## mask for normal
            mask_normal = np.where((per_bl_percentage_rn_0 > 0.9) | (per_bl_percentage_rn_0 < -0.9))
            ## if more than half of the detectors are normal
            if len(mask_normal[0]) > 0.5*len(mask_bg):
                bg_detectors_normal.append(bl)   
                drop_from_normal = True

## drop normal detector into transition first
    while drop_from_normal:
        S.log(f"some biaslines are still in normal state: {bg_detectors_normal}")
        percentage_rn_0 = bsa_0.R0/bsa_0.R_n_IV
        S.log("droping detectors from normal to transition")
        previous_dc_biases_dfn = S.get_tes_bias_bipolar_array()

        S.log(f"current tes bias voltages are: {previous_dc_biases_dfn}")
        for bl in bg_detectors_normal:
            mask_bg = np.where(bsa_0.bgmap==bl)[0]
            v_sc = iva.v_bias[iva.idxs[[mask_bg], 0]]
            v_norm = iva.v_bias[iva.idxs[[mask_bg], 1]]
            v_spread = np.nanmedian(v_norm) - np.nanmedian(v_sc)
            if previous_dc_biases_dfn[bl] > v_spread:
                previous_dc_biases_dfn[bl] =  previous_dc_biases_dfn[bl] - 0.5 * v_spread
            else:
                previous_dc_biases_dfn[bl] =  previous_dc_biases_dfn[bl] - 0.3 * v_spread 
                bg_detectors_normal.remove(bl)
                S.log(f'biasline {bl} is approching 0 voltage, a confirmation new IV curve is recommended')
        
        S.set_tes_bias_bipolar_array(previous_dc_biases_dfn)
        S.log(f"applying: {previous_dc_biases_dfn}")
        S.log(f"waiting 10s for dissipation of UFM local heating")
        time.sleep(10)
        bsa_0 = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)
        percentage_rn_0 = bsa_0.R0/bsa_0.R_n_IV
        drop_from_normal = False
        for bl in bg_detectors_normal:
            mask_bg = np.where(bsa_0.bgmap==bl)[0]
            per_bl_percentage_rn_0 = percentage_rn_0[mask_bg]
            ## mask for normal
            mask_normal = np.where((per_bl_percentage_rn_0 > 0.9) | (per_bl_percentage_rn_0 < -0.9))
            ## if more than half of the detectors are normal
            if len(mask_normal[0]) > 0.5*len(mask_bg):
                drop_from_normal = True
            else:
                bg_detectors_normal.remove(bl)
        for bl in bg_detectors_normal:
            if previous_dc_biases_dfn[bl] < 0.5:
                bg_detectors_normal.remove(bl)


                

    ##preparing for the 2nd step
    initial_dc_biases = S.get_tes_bias_bipolar_array()
    new_bias_voltage = initial_dc_biases.copy()




    for bl in bias_groups:
        mask_bg = np.where(bsa_0.bgmap==bl)[0]

        per_bl_percentage_rn_0 = percentage_rn_0[mask_bg]
            
        med_per_bl_percentage_rn = np.nanmedian(per_bl_percentage_rn_0)
        
        ## use the latest IV to decide the step size, might be tricky to get a good step size in field
        v_sc = iva.v_bias[iva.idxs[[mask_bg], 0]]
        v_norm = iva.v_bias[iva.idxs[[mask_bg], 1]]
        v_spread = np.nanmedian(v_norm) - np.nanmedian(v_sc)
        ## take 15% of the spread as the step size, so the next biasstep will be taken from 0.25 ~ 0.75 percent Rn after considering the spread
        delta_dc_bias = 0.15 * v_spread
        
        
        ## if the current bias point is below 50% Rn, increase V_biaa
        if med_per_bl_percentage_rn < 0.5:
            new_bias_voltage[bl] =  initial_dc_biases[bl] + delta_dc_bias
        if med_per_bl_percentage_rn >= 0.5:
            new_bias_voltage[bl] =  initial_dc_biases[bl] - delta_dc_bias

    S.log(f"applying new voltage for 2nd biasstep {new_bias_voltage}")
    S.set_tes_bias_bipolar_array(new_bias_voltage)  

    ## taking the second bias step
    S.log("taking the 2nd biasstep")
    bsa_1 = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)
    if make_plots:
        fig, ax = bias_steps.plot_Rfrac(bsa_1)
        fname = sdl.make_filename(S, '2nd_Rfrac.png', plot=True)
        fig.savefig(fname)
        S.pub.register_file(fname, 'biasstep_rfrac', format='png', plot=True)
        if not show_plots:
            plt.close(fig)

    ## check for script failue
    for bl in bias_groups:
        mask_bg = np.where(bsa_1.bgmap==bl)[0]
        per_bl_R0 = bsa_1.R0[mask_bg]
        if np.isnan(np.nanmedian(per_bl_R0)):
            ## the yield is very low, could it be the biasstep analysis script failed? repeat again
            repeat_biasstep = True
            S.log("the biasstep script might have failed, going to repeat the measurement")

    if repeat_biasstep:
        bsa_1 = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)
        repeat_biasstep = False

    percentage_rn_1 = bsa_1.R0/bsa_1.R_n_IV

    target_percentage_rn_array = target_percentage_rn * np.ones(len(percentage_rn_0))
    ## getting result from the previous 2
    v0 = np.zeros(bsa_0.bgmap.shape[0])
    for bl in range(12):
        mask_bl = np.where(bsa_0.bgmap==bl)[0]
        v0[mask_bl] = initial_dc_biases[bl]

    v1 = np.zeros(bsa_1.bgmap.shape[0])
    for bl in range(12):
        mask_bl = np.where(bsa_1.bgmap==bl)[0]
        v1[mask_bl] = new_bias_voltage[bl]   
    ## deciding the new bias point
    vbias_estimate_array = (v0*(target_percentage_rn_array-percentage_rn_1) + v1*(percentage_rn_0 - target_percentage_rn_array)) / (percentage_rn_0 - percentage_rn_1)

    vbias_estimate = initial_dc_biases.copy()
    for bl in bias_groups:
        mask_bg = np.where(bsa_0.bgmap==bl)[0]

        per_bl_vbias_estimate = vbias_estimate_array[mask_bg]
        med_per_bl_vbias_estimate = np.nanmedian(per_bl_vbias_estimate)
        if np.isnan(med_per_bl_vbias_estimate):
            med_per_bl_vbias_estimate = (np.nanmedian(v1[mask_bg]) + np.nanmedian(v0[mask_bg]))/2
        vbias_estimate[bl] = med_per_bl_vbias_estimate

    S.log("applying the new bias voltages")
    S.set_tes_bias_bipolar_array(vbias_estimate)  
    S.log(f"applying {vbias_estimate}")
    S.log("taking the final biasstep")
    bsa_2 = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)

    ## check for script failue
    for bl in bias_groups:
        mask_bg = np.where(bsa_2.bgmap==bl)[0]
        per_bl_R0 = bsa_2.R0[mask_bg]
        if np.isnan(np.nanmedian(per_bl_R0)):
            ## the yield is very low, could it be the biasstep analysis script failed? repeat again
            repeat_biasstep = True
            S.log("the biasstep script might have failed, going to repeat the measurement")

    if repeat_biasstep:
        bsa_2 = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)
        repeat_biasstep = False

    if make_plots:
        fig, ax = bias_steps.plot_Rfrac(bsa_2)
        fname = sdl.make_filename(S, 'rebiased_Rfrac.png', plot=True)
        fig.savefig(fname)
        S.pub.register_file(fname, 'biasstep_rfrac', format='png', plot=True)
        if not show_plots:
            plt.close(fig)


    S.log("confirming if the current result is close to the target")
    percentage_rn_confirm = bsa_2.R0/bsa_2.R_n_IV
    succeeded_bl = []
    un_succeeded_bl = []
    for bl in bias_groups:
        mask_bg = np.where(bsa_2.bgmap==bl)[0]
        per_bl_percentage_rn_confirm = np.nanmedian(percentage_rn_confirm[mask_bg])
        if per_bl_percentage_rn_confirm > target_percentage_rn - 0.05 and per_bl_percentage_rn_confirm < target_percentage_rn + 0.05:
            succeeded_bl.append(bl)
        else:
            un_succeeded_bl.append(bl)
    S.log(f"succeeded bl: {succeeded_bl}")
    S.log(f"unsucceeded bl: {un_succeeded_bl}")

    if len(un_succeeded_bl)==0:
        bsa_final = bsa_2
        vbias_estimate_final = vbias_estimate


    if len(un_succeeded_bl)!=0:
        extra_step_bias_voltage = vbias_estimate.copy()
        S.log("fine tunning unsucceeded bl")
        for bl in un_succeeded_bl:
            mask_bg = np.where(bsa_2.bgmap==bl)[0]
            per_bl_percentage_rn_confirm = percentage_rn_confirm[mask_bg]
            med_per_bl_percentage_confirm = np.nanmedian(per_bl_percentage_rn_confirm)

            ## use the latest IV to decide the step size, might be tricky to get a good step size in field
            v_sc = iva.v_bias[iva.idxs[[mask_bg], 0]]
            v_norm = iva.v_bias[iva.idxs[[mask_bg], 1]]
            v_spread = np.nanmedian(v_norm) - np.nanmedian(v_sc)
            ## take 15% of the spread as the step size, so the next biasstep will be taken from 0.25 ~ 0.75 percent Rn after considering the spread
            delta_dc_bias = 0.15 * v_spread
            ## if the current bias point is below 50% Rn, increase V_biaa
            if med_per_bl_percentage_confirm < 0.5:
                extra_step_bias_voltage[bl] =  vbias_estimate[bl] + delta_dc_bias
            if med_per_bl_percentage_confirm >= 0.5:
                extra_step_bias_voltage[bl] =  vbias_estimate[bl] - delta_dc_bias

        S.log(f"applying new voltage for extra step biasstep {extra_step_bias_voltage}")
        S.set_tes_bias_bipolar_array(extra_step_bias_voltage)

        ## taking the extra bias step
        S.log("taking the extra biasstep")
        bsa_extra = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)
        
        ## check for script failue
        for bl in bias_groups:
            mask_bg = np.where(bsa_extra.bgmap==bl)[0]
            per_bl_R0 = bsa_extra.R0[mask_bg]
            if np.isnan(np.nanmedian(per_bl_R0)):
            ## the yield is very low, could it be the biasstep analysis script failed? repeat again
                repeat_biasstep = True
                S.log("the biasstep script might have failed, going to repeat the measurement")

        if repeat_biasstep:
            bsa_extra = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)
            repeat_biasstep = False
        
        if make_plots:
            fig, ax = bias_steps.plot_Rfrac(bsa_extra)
            fname = sdl.make_filename(S, '3rd_Rfrac.png', plot=True)
            fig.savefig(fname)
            S.pub.register_file(fname, 'biasstep_rfrac', format='png', plot=True)
            if not show_plots:
                plt.close(fig)

        percentage_rn_extra = bsa_extra.R0/bsa_extra.R_n_IV

        ## getting result from the previous 2
        v0 = np.zeros(bsa_2.bgmap.shape[0])
        for bl in range(12):
            mask_bl = np.where(bsa_2.bgmap==bl)[0]
            v0[mask_bl] = vbias_estimate[bl]

        v1 = np.zeros(bsa_extra.bgmap.shape[0])
        for bl in range(12):
            mask_bl = np.where(bsa_extra.bgmap==bl)[0]
            v1[mask_bl] = extra_step_bias_voltage[bl]
        ## deciding the new bias point
        vbias_estimate_array_extra = (v0*(target_percentage_rn_array-percentage_rn_extra) + v1*(percentage_rn_confirm - target_percentage_rn_array)) / (percentage_rn_confirm - percentage_rn_extra)

        vbias_estimate_final = initial_dc_biases.copy()
        for bl in bias_groups:
            mask_bg = np.where(bsa_2.bgmap==bl)[0]

            per_bl_vbias_estimate = vbias_estimate_array_extra[mask_bg]
            med_per_bl_vbias_estimate = np.nanmedian(per_bl_vbias_estimate)
            if np.isnan(med_per_bl_vbias_estimate):
                med_per_bl_vbias_estimate = (np.nanmedian(v1[mask_bg]) + np.nanmedian(v0[mask_bg]))/2
            vbias_estimate_final[bl] = med_per_bl_vbias_estimate


        S.log("applying the new bias voltages")
        S.set_tes_bias_bipolar_array(vbias_estimate_final)
        S.log(f"applying {vbias_estimate_final}")
        S.log("taking the final biasstep")
        bsa_final = bias_steps.take_bias_steps(S, cfg, **bs_kwargs)
        if make_plots:
            fig, ax = bias_steps.plot_Rfrac(bsa_final)
            fname = sdl.make_filename(S, 'rebiased_Rfrac_2.png', plot=True)
            fig.savefig(fname)
            S.pub.register_file(fname, 'biasstep_rfrac', format='png', plot=True)
            if not show_plots:
                plt.close(fig)


    return bsa_final,vbias_estimate_final



import sodetlib as sdl
import numpy as np
import time
from scipy import signal
from tqdm.auto import trange
from pysmurf.client.base.smurf_control import SmurfControl
import matplotlib.pyplot as plt

def check_packet_loss(Ss, cfgs, dur=10, fr_khz=4, nchans=2000, slots=None, downsample_factor=1):
    """
    Takes a short G3 Stream on multiple slots simultaneously and checks for
    dropped samples. This function is strange since it requires simultaneous
    streaming on multiple slots to properly test, so it doesn't follow the
    standard sodetlib function / data format.

    Args
    -----
    Ss : dict[SmurfController]1
        Dict of pysmurf instances where the key is the slot-number
    cfgs : dict[DetConfig]
        Dict of DetConfigs where the key is the slot number
    dur : float
        Duration of data stream (sec)
    fr_khz : float
        Frequency of FR rate (khz)
    nchans : int
        Number of channels to stream
    slots : list
        Which slots to stream data on. If None, will stream on all slots in the
        Ss object.
    downsample_factor : int
        Downsample factor to use

    Returns
    --------
    ams : dict[AxisManagers]
        Dict of axis managers indexed by slot-number
    res : dict
        Dict where the key is the slot number, and the values are dicts
        containing frame counters, number of dropped frames, etc.
    """
    if slots is None:
        slots = Ss.keys()

    for s in slots:
        S = Ss[s]
        S.flux_ramp_setup(fr_khz, 0.4, band=0)
        sdl.stream_g3_on(
            S, channel_mask=np.arange(nchans), downsample_factor=downsample_factor,
            subtype='check_packet_loss'
        )

    time.sleep(dur)
    
    sids = {}
    for s in slots:
        sids[s] = sdl.stream_g3_off(Ss[s])

    ams = {}
    for s, sid in sids.items():
        ams[s] = sdl.load_session(cfgs[s].stream_id, sid, no_signal=True)

    res = {}
    for s, am in ams.items():
        dropped_samps = np.sum(np.diff(am.primary['FrameCounter']) - 1)
        total_samps = len(am.primary['FrameCounter'])
        res[s] = {
            'sid': sids[s],
            'meta': sdl.get_metadata(Ss[s], cfgs[s]),
            'frame_counter': am.primary['FrameCounter'],
            'dropped_samples': dropped_samps,
            'dropped_frac': dropped_samps / total_samps,
        }

    return ams, res

@sdl.set_action()
def measure_bias_line_resistances(
    S: SmurfControl, cfg, vstep=0.001, bgs=None, sleep_time=2.0):
    """
    Function to measure the bias line resistance and high-low-current-ratio for
    each bias group. This needs to be run with the smurf hooked up to the
    cryostat and the detectors superconducting, and a bgmap should exist.

    Args
    -------
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        Det Config instance
    vstep : float
        Voltage step size (in low-current-mode volts)
    bgs : list
        Bias lines to measure. Will default to active bias lines
    sleep_time : float
        Time to wait at each step.
    """
    if bgs is None:
        bgs = cfg.dev.exp['active_bgs']
    bgs = np.atleast_1d(bgs)
    bgmap = np.load(cfg.dev.exp['bgmap_file'], allow_pickle=True).item()
    bg_m = np.asarray(sum([bgmap['bgmap'] == bg for bg in bgs]), dtype=bool)

    vbias = S.get_tes_bias_bipolar_array()
    vb_low = vbias.copy()
    vb_low[bgs] = 0
    vb_high = vbias.copy()
    vb_high[bgs] = vstep
    segs = []

    S.set_tes_bias_bipolar_array(vb_low)
    sdl.set_current_mode(S, bgs, 0, const_current=False)

    def take_step(bias_arr, sleep_time, wait_time=0.2):
        S.set_tes_bias_bipolar_array(bias_arr)
        time.sleep(wait_time)
        t0 = time.time()
        time.sleep(sleep_time)
        t1 = time.time()
        return (t0, t1)

    sdl.stream_g3_on(S, subtype='measure_bias_line_resistance')
    time.sleep(0.5)

    segs.append(take_step(vb_low, sleep_time, wait_time=0.5))
    segs.append(take_step(vb_high, sleep_time, wait_time=0.5))

    S.set_tes_bias_bipolar_array(vb_low)
    time.sleep(0.5)
    sdl.set_current_mode(S, bgs, 1, const_current=False)

    segs.append(take_step(vb_low, sleep_time, wait_time=0.05))
    segs.append(take_step(vb_high, sleep_time, wait_time=0.05))

    sid = sdl.stream_g3_off(S)

    am = sdl.load_session(cfg.stream_id, sid)
    ts = am.timestamps
    sigs = []
    for (t0, t1) in segs:
        m = (t0 < ts) & (ts < t1)
        sigs.append(np.mean(am.signal[bg_m][:, m], axis=1) * S.pA_per_phi0 / (2*np.pi))

    Rbl_low = vstep / (np.abs(sigs[1] - sigs[0]) * 1e-12)
    Rbl_high = vstep / (np.abs(sigs[3] - sigs[2]) * 1e-12)
    high_low_ratio = Rbl_low / Rbl_high

    cfg.dev.exp['bias_line_resistance'] = np.nanmedian(Rbl_low)
    cfg.dev.exp['high_low_current_ratio'] = np.nanmedian(high_low_ratio)
    cfg.dev.update_file()

    path = sdl.make_filename(S, 'measure_bias_line_info')
    data = {
        'Rbl_low_all': Rbl_low,
        'Rbl_high_all': Rbl_high,
        'high_low_ratio_all': high_low_ratio,
        'bias_line_resistance': np.nanmedian(Rbl_low),
        'high_current_mode_resistance': np.nanmedian(Rbl_high),
        'high_low_ratio': np.nanmedian(high_low_ratio),
        'sid': sid,
        'vstep': vstep,
        'bgs': bgs,
        'meta': sdl.get_metadata(S, cfg),
        'segs': segs,
        'sigs': sigs,
        'path': path,
    }
    np.save(path, data, allow_pickle=True)
    S.pub.register_file(path, 'bias_line_resistances', format='npy')

    return am, data

def setup_fixed_tones(S, cfg, tones_per_band=256, bands=None, jitter=0.5,
                      tone_power=None):
    """
    Enables many fixed tones across a selection of bands.

    Args
    ----------
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        Det config instance
    tones_per_band : int
        Number of fixed tones to create in each band
    bands : int, list[int]
        Bands to set fixed tones in. Defaults to all 8.
    jitter : float
        Noise [Mhz] to add to the center freq so that fixed tones are not
        equispaced.
    tone_power : int
        Tone power of fixed tones.
    """
    if bands is None:
        bands = np.arange(8)
    bands = np.atleast_1d(bands)

    chans_per_band = S.get_number_channels()
    for band in bands:
        S.log(f"Setting fixed tones for band {band}")
        if tone_power is None:
            tone_power = cfg.dev.bands[band]['tone_power']
        
        sbs = np.linspace(0, chans_per_band, tones_per_band, dtype=int, endpoint=False)
        asa = np.zeros_like(S.get_amplitude_scale_array(band))
        asa[sbs] = tone_power
        S.set_amplitude_scale_array(band, asa)
        S.set_center_frequency_array(band, np.random.uniform(-jitter/2, jitter/2, chans_per_band))
        S.set_feedback_enable_array(band, np.zeros(chans_per_band, dtype=int))

def get_noise_dBcHz(S, band, chan, nsamp=2**20, nperseg=2**16,
                    noise_freq=30e3, noise_bw=100):
    """
    Takes debug data and measures I/Q noise in dBc/Hz.

    Args
    -----
    S : SmurfControl
        Pysmurf instance
    band : int
        Smurf band
    chan : int
        Smurf chan
    nsamp : int
        Number of samples to take
    nperseg : int
        Nperseg to use when creating the psd
    noise_freq : float
        Freq to measure the readout noise at.
    noise_bw : float
        Frequency bandwidth over which the noise median will be taken

    Returns
    ---------
    noise_i : float
        Noise of the I data stream in dBc/Hz
    noise_q : float
        Noise of the Q data stream in dBc/Hz
    """
    fsamp = S.get_channel_frequency_mhz() * 1e6
    
    sig_i, sig_q, _ = S.take_debug_data(band, channel=chan, rf_iq=True, nsamp=nsamp)
    datfile = S.get_streamdatawriter_datafile()

    fs, pxxi = signal.welch(sig_i, fs=fsamp, nperseg=nperseg)
    fs, pxxq = signal.welch(sig_q, fs=fsamp, nperseg=nperseg)

    magfac = np.mean(sig_q)**2 + np.mean(sig_i)**2
    pxxi_dbc = 10. * np.log10(pxxi/magfac)
    pxxq_dbc = 10. * np.log10(pxxq/magfac)

    f0, f1 = noise_freq - noise_bw/2, noise_freq + noise_bw/2
    m = (f0 < fs) & (fs < f1)
    noise_i = np.nanmedian(pxxi_dbc[m])
    noise_q = np.nanmedian(pxxq_dbc[m])

    return noise_i, noise_q, datfile
    

def plot_fixed_tone_loopback(res):
    """Plot results from fixed_tone_loopback"""
    fig, ax = plt.subplots()
    fig.patch.set_facecolor('white')

    ax.plot(res['freqs'], res['noise_q'], '.', alpha=0.8)
    ax.plot(res['freqs'], res['noise_i'], '.', alpha=0.8)
    txt = '\n'.join([
        f"num tones: {len(res['ft_freqs_all'])}",
        f"tone power: {res['tone_power']}",
        f"UC att: {res['att_uc']}",
        f"DC att: {res['att_dc']}",
    ])
    ax.text(0.05, 0.8, txt, transform=ax.transAxes,
            bbox=dict(fc='white', alpha=0.6))
    ax.set_xlabel("Freq [MHz]")
    ax.set_ylabel("Noise [dBc/Hz]")
    crate_id = res['meta']['crate_id']
    slot = res['meta']['slot']
    ax.set_title(f"Crate {crate_id}, Slot {slot}")

    return fig, ax


def fixed_tone_loopback(
        S, cfg, bands=None, tones_per_band=256, meas_chans_per_band=5,
        setup_tones=False, tone_power=12, show_pb=True, noise_freq=30e3,
        noise_bw=100, att_uc=None, att_dc=None):
    """
    Runs QC test to check noise levels across band with many fixed tones enabled.
    
    Args
    ------
    S : SmurfControl
        Pysmurf instance
    cfg : DetConfig
        Det config instance
    bands : int, list[int]
        Bands to run on. Default is all 8
    tones_per_band : int
        Number of fixed tones to enable per band. This defaults to 256, which
        is 2048 tones total with all 8 bands enabled.
    meas_chans_per_band : int
        Number of channels per band to measure readout noise
    setup_fixed_tones : bool
        If true, will set up fixed tones across specified bands. If false,
        this will skip the setup and assume tones are already set up.
    tone_power : int
        Tone power to use
    show_pb : bool
        If True will show a progress bar
    noise_freq : float
        Target frequency to measure the readout noise
    noise_bw : float
        Frequency bandwidth over which the noise median will be taken
    att_uc : int
        UC atten. If not set, will use current uc atten.
    att_dc : int
        DC atten. If not set, will use current dc atten
    """
    if bands is None:
        bands = np.arange(8)
    bands = np.atleast_1d(bands)

    if att_uc is None:
        att_uc = S.get_att_uc(bands[0])
    else:
        for b in bands:
            S.set_att_uc(b, att_uc)

    if att_dc is None:
        att_dc = S.get_att_dc(bands[0])
    else:
        for b in bands:
            S.set_att_dc(b, att_dc)

    if setup_tones:
        setup_fixed_tones(S, cfg, tones_per_band=tones_per_band, bands=bands,
                          tone_power=tone_power)
    else: # Just update the tone power
        S.log(f"Setting tone power to {tone_power} for bands {bands}...")
        for band in bands:
            asa = S.get_amplitude_scale_array(band)
            if tone_power in np.unique(asa):
                continue
            asa[asa != 0] = tone_power
            S.set_amplitude_scale_array(band, asa)
    
    meas_bands = []
    meas_chans = []
    meas_freqs = []
    ft_chans_all = []
    ft_freqs_all = []
    ft_bands_all = []
    S.log("Finding fixed tones and meas_channels")
    for band in bands:
        freqs = S.get_center_frequency_array(band) \
                 + S.get_tone_frequency_offset_mhz(band) \
                 + S.get_band_center_mhz(band)

        ft_chans = np.where(S.get_amplitude_scale_array(band))[0]
        freqs = freqs[ft_chans]
        sort_idx = np.argsort(freqs)
        meas_idx = np.unique(np.round(np.linspace(
            0, len(ft_chans) - 1, meas_chans_per_band)).astype(int))
        # We want to sort first so meas_chans are evenly distributed across
        # freq space
        meas_chans.append(ft_chans[sort_idx][meas_idx])
        meas_bands.append([band for _ in meas_idx])
        meas_freqs.append(freqs[sort_idx][meas_idx])
        ft_chans_all.append(ft_chans)
        ft_bands_all.append([band for _ in ft_chans])
        ft_freqs_all.append(freqs)
    meas_bands = np.hstack(meas_bands)
    meas_chans = np.hstack(meas_chans)
    meas_freqs = np.hstack(meas_freqs)
    ft_chans_all = np.hstack(ft_chans_all)
    ft_bands_all = np.hstack(ft_freqs_all)
    ft_freqs_all = np.hstack(ft_freqs_all)

    noise_i = np.full_like(meas_freqs, np.nan)
    noise_q = np.full_like(meas_freqs, np.nan)
    datfiles = []
    for i in trange(len(meas_bands), disable=not(show_pb)):
        b, c = meas_bands[i], meas_chans[i]
        S.log(f"Band {b}, Chan {c}")
        try:
            noise_i[i], noise_q[i], _ = get_noise_dBcHz(
                S, b, c, noise_freq=noise_freq,
                noise_bw=noise_bw)
        except IndexError as e:
            S.log(f"Take Data failed...\n{e}")
            S.log("Skipping channel")
        datfiles.append(S.get_streamdatawriter_datafile())

    fname = sdl.make_filename(S, 'fixed_tone_loopback.npy')
    res = dict(
        meta=sdl.get_metadata(S, cfg),
        bands=meas_bands, channels=meas_chans, freqs=meas_freqs,
        noise_i=noise_i, noise_q=noise_q,
        ft_bands_all=ft_bands_all, ft_channels_all=ft_chans_all,
        ft_freqs_all=ft_freqs_all, noise_freq=noise_freq,
        att_uc=att_uc, att_dc=att_dc, tone_power=tone_power,
        datfiles=datfiles
    )
    np.save(fname, res, allow_pickle=True)
    S.pub.register_file(fname, 'loopback', format='npy')


    fname = sdl.make_filename(S, 'fixed_tone_loopback.png', plot=True)
    fig, _ = plot_fixed_tone_loopback(res)
    fig.savefig(fname)
    S.pub.register_file(fname, 'loopback', format='png', plot=True)

    return res




from sotodlib.io import load_smurf
from sodetlib.util import Registers
from sodetlib.operations import uxm_setup
import os
import time

try:
    from pysmurf.client.util.pub import set_action
except:
    set_action = lambda : (lambda f : f)


def get_session_files(stream_id, session_id, idx=None,
                      base_dir='/data/so/timestreams'):
    """
    Gets a list of all files on the system corresponding to a given streaming
    session.

    Args
    ----
    stream_id : str
        stream_id for the stream you wish to load. Often this will be in
        cfg.stream_id
    session_id : int
        Session id corresonding with the stream session you wish to load.
        This is what is returned by the stream-taking functions.
    idx : int, list(int), optional
        Index of file you wish to load. Long streams are chunked into 10-minute
        files, and this parameter can used to isolate a smaller part of a long
        stream.
    base_dir : str
        Base directory where timestreams are stored. Defaults to
        /data/so/timestreams.
    """
    subdir = os.path.join(base_dir, str(session_id)[:5], stream_id)
    files = sorted([
        os.path.join(subdir, f) for f in os.listdir(subdir)
        if str(session_id) in f
    ])

    if idx is None:
        return files
    elif isinstance(idx, int):
        return files[idx]
    else:  # list of indexes
        return [files[i] for i in idx]

def load_session_status(stream_id, session_id, base_dir='/data/so/timestreams'):
    """
    Gets SmurfStatus object for a given stream.

    Args
    ----
    stream_id : str
        stream_id for the stream you wish to load. Often this will be in
        cfg.stream_id
    session_id : int
        Session id corresonding with the stream session you wish to load.
        This is what is returned by the stream-taking functions.
    base_dir : str
        Base directory where timestreams are stored. Defaults to
        /data/so/timestreams.
    """
    files = get_session_files(stream_id, session_id, base_dir=base_dir)
    if len(files) == 0:
        raise FileNotFoundError(
            f"Could not find files for {(stream_id, session_id)}"
        )
    return load_smurf.SmurfStatus.from_file(files[0])


def load_session(stream_id, session_id, idx=None,
                 base_dir='/data/so/timestreams', show_pb=False, **kwargs):
    """
    Loads a stream-session into an axis manager. Any additional keyword
    arguments will be passed to the ``load_smurf.load_file`` function.

    Args
    ----
    stream_id : str
        stream_id for the stream you wish to load. Often this will be in
        cfg.stream_id
    session_id : int
        Session id corresonding with the stream session you wish to load.
        This is what is returned by the stream-taking functions.
    idx : int, list(int), optional
        Index of file you wish to load. Long streams are chunked into 10-minute
        files, and this parameter can used to isolate a smaller part of a long
        stream.
    base_dir : str
        Base directory where timestreams are stored. Defaults to
        /data/so/timestreams.
    """
    files = get_session_files(stream_id, session_id, idx=idx,
                              base_dir=base_dir)
    if len(files) == 0:
        raise FileNotFoundError(
            f"Could not find files for {(stream_id, session_id)}"
        )
    am =  load_smurf.load_file(files, show_pb=show_pb, **kwargs)

    if 'ch_info' not in am._fields:
        am.wrap('ch_info', am.det_info.smurf)

    return am


@set_action()
def take_g3_data(S, dur, **stream_kw):
    """
    Takes data for some duration

    Args
    ----
    S : SmurfControl
        Pysmurf control object
    dur : float
        Duration to take data over

    Returns
    -------
    session_id : int
        Id used to read back stream data
    """
    stream_g3_on(S, **stream_kw)
    time.sleep(dur)
    return stream_g3_off(S, emulator=stream_kw.get('emulator', False))


@set_action()
def stream_g3_on(S, make_freq_mask=False, emulator=False, tag=None,
                 channel_mask=None, filter_wait_time=2, make_datfile=False,
                 downsample_factor=None, downsample_mode=None,
                 filter_disable=False, filter_order=None, filter_cutoff=None,
                 stream_type=None, subtype='stream', enable_compression=None):
    """
    Starts the G3 data-stream. Returns the session-id corresponding with the
    data stream.

    Args
    ----
    S : S
        Pysmurf control object
    make_freq_mask : bool, optional
        Tell pysmurf to write and register the current freq mask
    emulator : bool
        If True, will enable the emulator source data-generator. Defaults to
        False.
    tag : optional(string)
        If set, this will attach the tag to the g3 stream
    channel_mask : optional(list)
        List of channels to stream. If None, this will stream all channels with
        a probe tone
    filter_wait_time : float
        Seconds to wait after resetting the filter before writing data to disk
    make_datfile bool
        If True, will create a datfile
    downsample_factor : optional(int)
        Downsample factor. If None, this will be pulled from the device
        cfg.
    downsample_mode : optional(string)
        Downsample mode, can either be 'internal' or 'external'. If not set,
        this will be pulled from the device cfg.
    filter_disable : bool
        If true, will disable the downsample filter before streaming.
    filter_order : int
        Order of the downsample filter. Read from device config by default.
    filter_cutoff : float
        The cutoff frequency in Hz for the downsample filter.
        Read from device config by default.
    stream_type : optional(string)
        Type of stream. This should either be "obs" or "oper". If None,
        it will be determined based off of the subtype. ('stream', 'cmb', and
        'cal', will automatically be set to 'obs' type)
    subtype : optional(string)
        Stream subtype. This defaults to 'stream'.
    enable_compression : optional(bool)
        If true, will tell the smurf-streamer to compress data.

    Return
    -------
    session_id : int
        Id used to read back streamed data
    """
    cfg = S._sodetlib_cfg
    if stream_type is None:
        if subtype in ['stream', 'cmb', 'cal']:
            stream_type = 'obs'
        else:
            stream_type = 'oper'

    if enable_compression is None:
        enable_compression = cfg.dev.exp.get('enable_compression', True)

    tags = f'{stream_type},{subtype}'
    if tag is not None:
        tags += ',' + tag

    reg = Registers(S)

    reg.pysmurf_action.set(S.pub._action)
    reg.pysmurf_action_timestamp.set(S.pub._action_ts)
    reg.stream_tag.set(tags)

    reg.enable_compression.set(enable_compression)

    if downsample_mode is None:
        downsample_mode = cfg.dev.exp['downsample_mode']
    if downsample_factor is None:
        downsample_factor = cfg.dev.exp['downsample_factor']
    if filter_order is None:
        filter_order = cfg.dev.exp["downsample_filter"]["order"]
    if filter_cutoff is None:
        filter_cutoff = cfg.dev.exp["downsample_filter"]["cutoff_freq"]

    S.set_downsample_mode(downsample_mode)
    S.set_downsample_factor(downsample_factor)
    S.set_downsample_filter(filter_order, filter_cutoff)
    S.set_filter_disable(int(filter_disable))

    # ensure fixed tones are on
    uxm_setup.turn_on_fixed_tones(S, cfg)

    S.stream_data_on(make_freq_mask=make_freq_mask, channel_mask=channel_mask,
                     filter_wait_time=filter_wait_time, make_datafile=make_datfile)

    if emulator:
        reg.source_enable.set(1)
        S.set_stream_enable(1)

    reg.open_g3stream.set(1)

    # Sometimes it takes a bit for data to propogate through to the
    # streamer
    for _ in range(10):
        sess_id = reg.g3_session_id.get()
        if sess_id != 0:
            break
        time.sleep(0.5)
    S.log(f"Session id: {sess_id}")
    return sess_id


@set_action()
def stream_g3_off(S, emulator=False):
    """
    Stops the G3 data-stream. Returns the session-id corresponding with the
    data stream.

    Args
    ----
    S : S
        Pysmurf control object
    emulator : bool
        If True, will enable the emulator source data-generator. Defaults to
        False.

    Return
    -------
    session_id : int
        Id used to read back streamed data
    """
    reg = Registers(S)
    sess_id = reg.g3_session_id.get()
    if sess_id == 0:
        S.log("Session-id returned 0! Will try to obtain from file path")
        fpath = reg.g3_filepath.get()
        try:
            sess_id = int(os.path.basename(fpath).split('_')[0])
        except Exception:
            S.log("Could not extract session id from filepath! Setting to 0")
            sess_id = 0

    if emulator:
        reg.source_enable.set(0)

    S.set_stream_enable(0)
    S.stream_data_off()

    reg.open_g3stream.set(0)
    reg.pysmurf_action.set('')
    reg.pysmurf_action_timestamp.set(0)
    reg.stream_tag.set('')

    # Waits until file is closed out before returning
    S.log("Waiting for g3 file to close out")
    while reg.g3_session_id.get() != 0:
        time.sleep(0.5)

    return sess_id




from setuptools import setup, Extension, find_packages

import versioneer

setup (name = 'sodetlib',
       description='SODETLIB',
       packages=find_packages(),
       scripts=['hammers/jackhammer'],
       version=versioneer.get_version(),
       cmdclass=versioneer.get_cmdclass())
